import snakes
from snakes.utils.simul.httpd import *
from snakes.utils.simul.html import H
from snakes.utils.simul import logger as log
import multiprocessing, time, sys, os.path, signal, inspect, glob
import operator

class StateSpace (dict) :
    def __init__ (self, net) :
        self.net = net
        self.current = self.add(net.get_marking())
    def get (self) :
        return self[self.current]
    def add (self, marking) :
        if marking in self :
            return self[marking].num
        else :
            marking.num = len(self) / 2
            self[marking] = self[marking.num] = marking
            self.setmodes(marking.num)
            return marking.num
    def setmodes (self, state) :
        marking = self[state]
        self.net.set_marking(marking)
        marking.modes = []
        for trans in self.net.transition() :
            for mode in trans.modes() :
                marking.modes.append((trans, mode))
    def succ (self, state, mode) :
        marking = self[state]
        trans, binding = marking.modes[mode]
        self.net.set_marking(marking)
        trans.fire(binding)
        self.current = self.add(self.net.get_marking())
        return self.current
    def modes (self, state) :
        return self[state].modes

shutdown = multiprocessing.Event()
ping = multiprocessing.Event()

class Server (multiprocessing.Process) :
    def __init__ (self, httpd) :
        multiprocessing.Process.__init__(self)
        self.httpd = httpd
    def run (self) :
        try :
            self.httpd.serve_forever()
        except KeyboardInterrupt :
            pass
        finally :
            shutdown.set()

class WatchDog (multiprocessing.Process) :
    def __init__ (self, timeout=30) :
        multiprocessing.Process.__init__(self)
        self.timeout = timeout
    def run (self) :
        try :
            while True :
                if ping.wait(self.timeout) :
                    ping.clear()
                else :
                    log.info("client has gone", "simul")
                    break
        except KeyboardInterrupt :
            pass
        finally :
            shutdown.set()

class BaseSimulator (object) :
    def __init__ (self, net) :
        self.states = StateSpace(net)
    def getstate (self, state) :
        marking = self.states[state]
        places = ["%s = %s" % (H.span(place.name, class_="place"),
                               H.span(marking(place.name), class_="token"))
                  for place in sorted(self.states.net.place(),
                                      key=operator.attrgetter("name"))]
        modes = [{"state" : state,
                  "mode" : i,
                  "html" : "%s : %s" % (H.span(trans.name, class_="trans"),
                                        H.span(binding, class_="binding"))}
                  for i, (trans, binding) in enumerate(marking.modes)]
        return {"id" : state,
                "states" : [{"do" : "sethtml",
                             "select" : "#net",
                             "html" : H.i(self.states.net)},
                            {"do" : "settext",
                             "select" : "#state",
                             "text" : state},
                            {"do" : "setlist",
                             "select" : "#marking",
                             "items" : places},
                            ],
                "modes" : [{"select" : "#modes",
                            "items" : modes},
                           ],
                }
    def init (self, state=-1) :
        if state < 0 :
            state = self.states.current
        return {"state" : self.getstate(state)}
    def succ (self, state, mode) :
        state = self.states.succ(state, mode)
        return self.getstate(state)
    def init_help (self) :
        return {
            "#trace": {
                "title" : "Trace",
                "content" : "the states and transitions explored so far"
            },
            "#model" : {
                "title" : "Model",
                "content" : "the model being simulated"
            },
            "#alive .ui #ui-quit" : {
                "title" : "Stop",
                "content" : "stop the simulator (server side)"
            },
            "#alive .ui #ui-help" : {
                "title" : "Help",
                "content" : "show this help"
            },
            "#alive .ui #ui-about" : {
                "title" : "About",
                "content" : "show information about the simulator"
            },
        }

class BaseHTTPSimulator (Node) :
    def __init__ (self, net=None, port=8000, respatt=[], simulator=None) :
        self.res = {}
        dirs = {}
        for cls in reversed(inspect.getmro(self.__class__)[:-2]) :
            path = os.path.dirname(inspect.getsourcefile(cls))
            for pattern in respatt + ["resources/*/*.*",
                                      "resources/*.*"] :
                for res in glob.glob(os.path.join(path, pattern)) :
                    if os.path.isfile(res) :
                        with open(res) as infile :
                            self.res[res[len(path + "resources/")+1:]] = infile.read()
                    elif os.path.isdir(res) :
                        dirs[os.path.basename(res)] = DirNode(res)
                    else :
                        raise ValueError("invalid resource %r" % res)
        Node.__init__(self, r=ResourceNode(self.res, dirs))
        # create HTTP server
        self.port = port
        while True :
            try :
                httpd = HTTPServer(('', self.port), self)
            except :
                self.port += 1
            else :
                break
        self.server = Server(httpd)
        self.watchdog = WatchDog()
        # init data
        self.url = "http://127.0.0.1:%s/%s/" % (port, httpd.key)
        self._alive = self.res["alive.txt"].splitlines()
        self._ping = 0
        if simulator is None :
            self.simul = BaseSimulator(net)
        else :
            self.simul = simulator
    def start (self) :
        log.info("starting at %r" % self.url, "simul")
        shutdown.clear()
        ping.clear()
        self.server.start()
        self.watchdog.start()
    def wait (self) :
        try :
            shutdown.wait()
            log.info("preparing to shut down...", "simul")
            time.sleep(2)
        except KeyboardInterrupt :
            shutdown.set()
        log.info("shuting down...", "simul")
        sig = getattr(signal, "CTRL_C_EVENT",
                      getattr(signal, "SIGTERM", None))
        if sig is not None :
            if self.server.pid :
                os.kill(self.server.pid, sig)
            if self.watchdog.pid :
                os.kill(self.watchdog.pid, sig)
        log.info("bye!", "simul")
    def init_index (self) :
        return {"res" : "%sr" % self.url,
                "url" : self.url,
                "key" : self.server.httpd.key,
                "host" : "127.0.0.1",
                "port" : self.port,
                "about" : self.init_about(),
                "model" : self.init_model()}
    def init_model (self) :
        return self.res["model.html"]
    def init_about (self) :
        return self.res["about.html"]
    @http("text/html")
    def __call__ (self) :
        return self.res["index.html"] % self.init_index()
    def init_ui (self) :
        argv = H.code(" ".join(sys.argv))
        version = (H.ul(H.li(H.b("Python: "),
                             H.br.join(sys.version.splitlines())),
                        H.li(H.b("SNAKES: "), snakes.version)))
        return [{"label" : "Versions",
                 "id" : "ui-version",
                 "href" : "#",
                 "script" : "dialog(%r)" % version},
                {"label" : "Argv",
                 "id" : "ui-argv",
                 "href" : "#",
                 "script" : "dialog(%r)" % argv}]
    @http("application/json", state=int)
    def init (self, state=-1) :
        ret = {"ui" : self.init_ui(),
               "help" : self.simul.init_help()}
        ret.update(self.simul.init(state))
        return ret
    @http("application/json", state=int, mode=int)
    def succ (self, state, mode) :
        return self.simul.succ(state, mode)
    @http("text/plain")
    def ping (self) :
        ping.set()
        alive = self._alive[self._ping % len(self._alive)]
        self._ping += 1
        return alive
    @http("text/plain")
    def quit (self) :
        shutdown.set()
        return "Bye!"

if __name__ == "__main__" :
    import snakes.nets, webbrowser
    net = snakes.nets.loads(sys.argv[1])
    simul = BaseHTTPSimulator(net)
    simul.start()
    webbrowser.open(simul.url)
    simul.wait()



import sys, operator, inspect, re, collections
from snakes.utils.abcd import CompilationError, DeclarationError
from snakes.lang.abcd.parser import ast
from snakes.lang import unparse
import snakes.utils.abcd.transform as transform
from snakes.data import MultiSet
from snakes import *
from snakes.compat import *

class Decl (object) :
    OBJECT = "object"
    TYPE = "type"
    BUFFER = "buffer"
    SYMBOL = "symbol"
    CONST = "const"
    NET = "net"
    TASK = "task"
    IMPORT = "import"
    def __init__ (self, node, kind=None, **data) :
        self.node = node
        classname = node.__class__.__name__
        if kind is not None :
            self.kind = kind
        elif classname == "AbcdTypedef" :
            self.kind = self.TYPE
        elif classname == "AbcdBuffer" :
            self.kind = self.BUFFER
        elif classname == "AbcdSymbol" :
            self.kind = self.SYMBOL
        elif classname == "AbcdConst" :
            self.kind = self.CONST
        elif classname == "AbcdNet" :
            self.kind = self.NET
        elif classname == "AbcdTask" :
            self.kind = self.TASK
        elif classname in ("Import", "ImportFrom") :
            self.kind = self.IMPORT
        else :
            self.kind = self.OBJECT
        for key, val in data.items() :
            setattr(self, key, val)

class GetInstanceArgs (object) :
    """Bind arguments for a net instance
    """
    def __init__ (self, node) :
        self.argspec = []
        self.arg = {}
        self.buffer = {}
        self.net = {}
        self.task = {}
        seen = set()
        for a in node.args.args + node.args.kwonlyargs :
            if a.arg in seen :
                self._raise(CompilationError,
                            "duplicate argument %r" % a.arg)
            seen.add(a.arg)
            if a.annotation is None :
                self.argspec.append((a.arg, "arg"))
            else :
                self.argspec.append((a.arg, a.annotation.id))
    def __call__ (self, *args) :
        self.arg.clear()
        self.buffer.clear()
        self.net.clear()
        self.task.clear()
        for (name, kind), value in zip(self.argspec, args) :
            getattr(self, kind)[name] = value
        return self.arg, self.buffer, self.net, self.task

class Builder (object) :
    def __init__ (self, snk, path=[], up=None) :
        self.snk = snk
        self.path = path
        self.up = up
        self.env = {"True": Decl(None, kind=Decl.CONST, value=True),
                    "False": Decl(None, kind=Decl.CONST, value=False),
                    "None": Decl(None, kind=Decl.CONST, value=None),
                    "dot": Decl(None, kind=Decl.CONST, value=self.snk.dot),
                    "BlackToken": Decl(None, kind=Decl.TYPE,
                                       type=self.snk.Instance(self.snk.BlackToken))}
        self.stack = []
        if up :
            self.globals = up.globals
        else :
            self.globals = snk.Evaluator(dot=self.snk.dot,
                                         BlackToken=self.snk.BlackToken)
        self.instances = MultiSet()
    # utilities
    def _raise (self, error, message) :
        """raise an exception with appropriate location
        """
        if self.stack :
            try :
                pos = "[%s:%s]: " % (self.stack[-1].lineno,
                                     self.stack[-1].col_offset)
            except :
                pos = ""
        else :
            pos = ""
        raise error(pos+message)
    def _eval (self, expr, *largs, **kwargs) :
        env = self.globals.copy()
        if isinstance(expr, ast.AST) :
            expr = unparse(expr)
        return env(expr, dict(*largs, **kwargs))
    # declarations management
    def __setitem__ (self, name, value) :
        if name in self.env :
            self._raise(DeclarationError, "duplicated declaration of %r" % name)
        self.env[name] = value
    def __getitem__ (self, name) :
        if name in self.env :
            return self.env[name]
        elif self.up is None :
            self._raise(DeclarationError, "%r not declared" % name)
        else :
            return self.up[name]
    def __contains__ (self, name) :
        if name in self.env :
            return True
        elif self.up is None :
            return False
        else :
            return name in self.up
    def goto (self, name) :
        if name in self.env :
            return self
        elif self.up is None :
            self._raise(DeclarationError, "%r not declared" % name)
        else :
            return self.up.goto(name)
    def get_buffer (self, name) :
        if name not in self :
            self._raise(DeclarationError,
                        "buffer %r not declared" % name)
        decl = self[name]
        if decl.kind != Decl.BUFFER :
            self._raise(DeclarationError,
                        "%r declared as %s but used as buffer"
                        % (name, decl.kind))
        elif decl.capacity is not None :
            pass
            #self._raise(NotImplementedError, "capacities not (yet) supported")
        return decl
    def get_net (self, name) :
        if name not in self :
            self._raise(DeclarationError,
                        "net %r not declared" % name)
        decl = self[name]
        if decl.kind != Decl.NET :
            self._raise(DeclarationError,
                        "%r declared as %s but used as net"
                        % (name, decl.kind))
        return decl
    def get_task (self, name) :
        if name not in self :
            self._raise(DeclarationError,
                        "task %r not declared" % name)
        decl = self[name]
        if decl.kind != Decl.TASK :
            self._raise(DeclarationError,
                        "%r declared as %s but used as task"
                        % (name, decl.kind))
        return decl
    # main compiler entry point
    def build (self, node, prefix="", fallback=None) :
        self.stack.append(node)
        if prefix :
            prefix += "_"
        method = "build_" + prefix + node.__class__.__name__
        visitor = getattr(self, method, fallback or self.build_fail)
        try :
            return visitor(node)
        finally :
            self.stack.pop(-1)
    def build_fail (self, node) :
        self._raise(CompilationError, "do not know how to compile %s"
                    % node.__class__.__name__)
    def build_arc (self, node) :
        return self.build(node, "arc", self.build_arc_expr)
    # specification
    def build_AbcdSpec (self, node) :
        for decl in node.context :
            self.build(decl)
        tasks = [self._build_TaskNet(decl.node)
                 for name, decl in self.env.items()
                 if decl.kind == Decl.TASK and decl.used]
        net = reduce(operator.or_, tasks, self.build(node.body))
        # set local buffers marking, and hide them
        for name, decl in ((n, d) for n, d in self.env.items()
                           if d.kind == Decl.BUFFER) :
            status = self.snk.buffer(name)
            for place in net.status(status) :
                place = net.place(place)
                try :
                    place.reset(decl.marking)
                except ValueError as err :
                    self._raise(CompilationError,
                                "invalid initial marking (%s)" % err)
                if decl.capacity is None :
                    cap = None
                else :
                    #cap = [c.n if c else None for c in decl.capacity]
                    # TODO: accept more than integers as capacities
                    cap = []
                    for c in decl.capacity :
                        if c is None :
                            cap.append(None)
                        else :
                            try :
                                cap.append(self._eval(c))
                            except :
                                err = sys.exc_info()[1]
                                self._raise(CompilationError,
                                            "could not evaluate %r, %s"
                                            % (unparse(c), err))
                place.label(path=self.path,
                            capacity=cap)
                # TODO: check capacity
            net.hide(status)
        if self.up is None :
            # set entry marking
            for place in net.status(self.snk.entry) :
                net.place(place).reset(self.snk.dot)
            # rename nodes
            self._rename_nodes(net)
            # copy global declarations
            net.globals.update(self.globals)
            # add info about source file
            net.label(srcfile=str(node.st.text.filename))
            # add assertions
            net.label(asserts=node.asserts)
        return net
    def _build_TaskNet (self, node) :
        self._raise(NotImplementedError, "tasks not (yet) supported")
    def _rename_nodes (self, net) :
        # generate unique names
        total = collections.defaultdict(int)
        count = collections.defaultdict(int)
        def ren (node) :
            if net.has_transition(node.name) :
                status = node.label("srctext")
            else :
                if node.status == self.snk.entry :
                    status = "e"
                elif node.status == self.snk.internal :
                    status = "i"
                elif node.status == self.snk.exit :
                    status = "x"
                else :
                    status = node.label("buffer")
            name = ".".join(node.label("path") + [status])
            if total[name] > 1 :
                count[name] += 1
                name = "%s#%s" % (name, count[name])
            return name
        # count occurrences of each name base
        _total = collections.defaultdict(int)
        for node in net.node() :
            _total[ren(node)] += 1
        total = _total
        # rename nodes using a depth-first traversal
        done = set(net.status(self.snk.entry))
        todo = [net.node(n) for n in done]
        while todo :
            node = todo.pop(-1)
            new = ren(node)
            if new != node.name :
                net.rename_node(node.name, new)
            done.add(new)
            for n in net.post(new) - done :
                todo.append(net.node(n))
                done.add(n)
        # rename isolated nodes
        for letter, method in (("p", net.place), ("t", net.transition)) :
            for node in method() :
                if node.name not in done :
                    net.rename_node(node.name, ren(node))
    # declarations
    def build_AbcdTypedef (self, node) :
        """
        >>> import snakes.nets
        >>> b = Builder(snakes.nets)
        >>> b.build(ast.AbcdTypedef(name='number', type=ast.UnionType(types=[ast.NamedType(name='int'), ast.NamedType(name='float')])))
        >>> b.env['number'].type
        (Instance(int) | Instance(float))
        >>> b.build(ast.ImportFrom(module='inspect', names=[ast.alias(name='isbuiltin')]))
        >>> b.build(ast.AbcdTypedef(name='builtin', type=ast.NamedType(name='isbuiltin')))
        >>> b.env['builtin'].type
        TypeCheck(inspect.isbuiltin)
        """
        self[node.name] = Decl(node, type=self.build(node.type))
    def build_AbcdBuffer (self, node) :
        self[node.name] = Decl(node,
                               type=self.build(node.type),
                               capacity=node.capacity,
                               marking=self._eval(node.content))
    def build_AbcdSymbol (self, node) :
        for name in node.symbols :
            value = self.snk.Symbol(name, False)
            self[name] = Decl(node, value=value)
            self.globals[name] = value
    def build_AbcdConst (self, node) :
        value = self._eval(node.value)
        self[node.name] = Decl(node, value=value)
        self.globals[node.name] = value
    def build_AbcdNet (self, node) :
        self[node.name] = Decl(node, getargs=GetInstanceArgs(node))
    def build_AbcdTask (self, node) :
        self._raise(NotImplementedError, "tasks not (yet) supported")
        self[node.name] = Decl(node, used=False)
    def build_Import (self, node) :
        for alias in node.names :
            self[alias.asname or alias.name] = Decl(node)
        self.globals.declare(unparse(node))
    def build_ImportFrom (self, node) :
        self.build_Import(node)
    # processes
    def build_AbcdAction (self, node) :
        if node.guard is True :
            return self._build_True(node)
        elif node.guard is False :
            return self._build_False(node)
        else :
            return self._build_action(node)
    def _build_True (self, node) :
        net = self.snk.PetriNet("true")
        e = self.snk.Place("e", [], self.snk.tBlackToken,
                           status=self.snk.entry)
        e.label(path=self.path)
        net.add_place(e)
        x = self.snk.Place("x", [], self.snk.tBlackToken,
                           status=self.snk.exit)
        x.label(path=self.path)
        net.add_place(x)
        t = self.snk.Transition("t")
        t.label(srctext=node.st.source(),
                srcloc=(node.st.srow, node.st.scol,
                        node.st.erow, node.st.ecol),
                path=self.path)
        net.add_transition(t)
        net.add_input("e", "t", self.snk.Value(self.snk.dot))
        net.add_output("x", "t", self.snk.Value(self.snk.dot))
        return net
    def _build_False (self, node) :
        net = self.snk.PetriNet("false")
        e = self.snk.Place("e", [], self.snk.tBlackToken,
                           status=self.snk.entry)
        e.label(path=self.path)
        net.add_place(e)
        x = self.snk.Place("x", [], self.snk.tBlackToken,
                           status=self.snk.exit)
        x.label(path=self.path)
        net.add_place(x)
        return net
    def _build_action (self, node) :
        net = self.snk.PetriNet("flow")
        e = self.snk.Place("e", [], self.snk.tBlackToken,
                           status=self.snk.entry)
        e.label(path=self.path)
        net.add_place(e)
        x = self.snk.Place("x", [], self.snk.tBlackToken,
                           status=self.snk.exit)
        x.label(path=self.path)
        net.add_place(x)
        t = self.snk.Transition("t", self.snk.Expression(unparse(node.guard)),
                                status=self.snk.tick("action"))
        t.label(srctext=node.st.source(),
                srcloc=(node.st.srow, node.st.scol,
                        node.st.erow, node.st.ecol),
                path=self.path)
        net.add_transition(t)
        net.add_input("e", "t", self.snk.Value(self.snk.dot))
        net.add_output("x", "t", self.snk.Value(self.snk.dot))
        net = reduce(operator.or_, [self.build(a) for a in node.accesses],
                     net)
        net.hide(self.snk.tick("action"))
        return net
    def build_AbcdFlowOp (self, node) :
        return self.build(node.op)(self.build(node.left),
                                   self.build(node.right))
    def _get_instance_arg (self, arg) :
        if arg.__class__.__name__ == "Name" and arg.id in self :
            return self[arg.id]
        else :
            try :
                self._eval(arg)
            except :
                self._raise(CompilationError,
                            "could not evaluate argument %r"
                            % arg.st.source())
            return arg
    def build_AbcdInstance (self, node) :
        if node.net not in self :
            self._raise(DeclarationError, "%r not declared" % node.net)
        elif node.starargs :
            self._raise(CompilationError, "* argument not allowed here")
        elif node.kwargs :
            self._raise(CompilationError, "** argument not allowed here")
        decl = self[node.net]
        if decl.kind != Decl.NET :
            self._raise(DeclarationError,
                        "%r declared as %s but used as net"
                        % (name, decl.kind))
        # unpack args
        posargs, kwargs = [], {}
        for arg in node.args :
            posargs.append(self._get_instance_arg(arg))
        for kw in node.keywords :
            kwargs[kw.arg] = self._get_instance_arg(kw.value)
        # bind args
        try :
            args, buffers, nets, tasks = decl.getargs(*posargs, **kwargs)
        except TypeError :
            c, v, t = sys.exc_info()
            self._raise(CompilationError, str(v))
        for d, kind in ((buffers, Decl.BUFFER),
                        (nets, Decl.NET),
                        (tasks, Decl.TASK)) :
            for k, v in d.items() :
                if v.kind != kind :
                    self._raise(DeclarationError,
                                "%r declared as %s but used as %s"
                                % (k, v.kind, kind))
                d[k] = v.node.name
        # build sub-net
        binder = transform.ArgsBinder(args, buffers, nets, tasks)
        spec = binder.visit(decl.node.body)
        if node.asname :
            name = str(node.asname)
        else :
            name = node.st.source()
        if name in self.instances :
            name = "%s#%s" % (name, self.instances(name))
        self.instances.add(name)
        path = self.path + [name]
        builder = self.__class__(self.snk, path, self)
        net = builder.build(spec)
        src = (node.st.source(),
               node.st.srow, node.st.scol,
               node.st.erow, node.st.ecol)
        for trans in net.transition() :
            try :
                lbl = trans.label("instances")
                trans.label(instances=[src] + lbl)
            except KeyError :
                trans.label(instances=[src])
        for place in net.place() :
            if place.status == self.snk.Status(None) :
                try :
                    lbl = place.label("instances")
                    place.label(instances=[src] + lbl)
                except KeyError :
                    place.label(instances=[src])
        return net
    # control flow operations
    def build_Sequence (self, node) :
        return self.snk.PetriNet.__and__
    def build_Choice (self, node) :
        return self.snk.PetriNet.__add__
    def build_Parallel (self, node) :
        return self.snk.PetriNet.__or__
    def build_Loop (self, node) :
        return self.snk.PetriNet.__mul__
    # accesses :
    def build_SimpleAccess (self, node) :
        decl = self.get_buffer(node.buffer)
        net = self.snk.PetriNet("access")
        net.add_transition(self.snk.Transition("t", status=self.snk.tick("action")))
        b = self.snk.Place(str(node.buffer), [], decl.type,
                           status=self.snk.buffer(node.buffer))
        b.label(path=self.path,
                buffer=str(node.buffer),
                srctext=decl.node.st.source(),
                srcloc=(decl.node.st.srow, decl.node.st.scol,
                        decl.node.st.erow, decl.node.st.ecol))
        net.add_place(b)
        self.build(node.arc)(net, node.buffer, "t", self.build_arc(node.tokens))
        return net
    def build_FlushAccess (self, node) :
        decl = self.get_buffer(node.buffer)
        net = self.snk.PetriNet("access")
        net.add_transition(self.snk.Transition("t", status=self.snk.tick("action")))
        b = self.snk.Place(str(node.buffer), [], decl.type,
                           status=self.snk.buffer(node.buffer))
        b.label(path=self.path,
                buffer=str(node.buffer),
                srctext=decl.node.st.source(),
                srcloc=(decl.node.st.srow, decl.node.st.scol,
                        decl.node.st.erow, decl.node.st.ecol))
        net.add_place(b)
        net.add_input(node.buffer, "t", self.snk.Flush(node.target))
        return net
    def build_SwapAccess (self, node) :
        decl = self.get_buffer(node.buffer)
        net = self.snk.PetriNet("access")
        net.add_transition(self.snk.Transition("t", status=self.snk.tick("action")))
        b = self.snk.Place(node.buffer, [], decl.type,
                           status=self.snk.buffer(node.buffer))
        b.label(path=self.path,
                buffer=str(node.buffer),
                srctext=decl.node.st.source(),
                srcloc=(decl.node.st.srow, decl.node.st.scol,
                        decl.node.st.erow, decl.node.st.ecol))
        net.add_place(b)
        net.add_input(node.buffer, "t", self.build_arc(node.target))
        net.add_output(node.buffer, "t", self.build_arc(node.tokens))
        return net
    def build_Spawn (self, node) :
        self._raise(NotImplementedError, "tasks not (yet) supported")
    def build_Wait (self, node) :
        self._raise(NotImplementedError, "tasks not (yet) supported")
    def build_Suspend (self, node) :
        self._raise(NotImplementedError, "tasks not (yet) supported")
    def build_Resume (self, node) :
        self._raise(NotImplementedError, "tasks not (yet) supported")
    # arc labels
    def build_arc_Name (self, node) :
        if node.id in self :
            decl = self[node.id]
            if decl.kind in (Decl.CONST, Decl.SYMBOL) :
                return self.snk.Value(decl.value)
        return self.snk.Variable(node.id)
    def build_arc_Num (self, node) :
        return self.snk.Value(node.n)
    def build_arc_Str (self, node) :
        return self.snk.Value(node.s)
    def build_arc_Tuple (self, node) :
        return self.snk.Tuple([self.build_arc(elt) for elt in node.elts])
    def build_arc_expr (self, node) :
        return self.snk.Expression(unparse(node))
    # arcs
    def build_Produce (self, node) :
        def arc (net, place, trans, label) :
            net.add_output(place, trans, label)
        return arc
    def build_Test (self, node) :
        def arc (net, place, trans, label) :
            net.add_input(place, trans, self.snk.Test(label))
        return arc
    def build_Consume (self, node) :
        def arc (net, place, trans, label) :
            net.add_input(place, trans, label)
        return arc
    def build_Fill (self, node) :
        def arc (net, place, trans, label) :
            net.add_output(place, trans, self.snk.Flush(str(label)))
        return arc
    # types
    def build_UnionType (self, node) :
        return reduce(operator.or_, (self.build(child)
                                     for child in node.types))
    def build_IntersectionType (self, node) :
        return reduce(operator.and_, (self.build(child)
                                      for child in node.types))
    def build_CrossType (self, node) :
        return self.snk.CrossProduct(*(self.build(child)
                                       for child in node.types))
    def build_ListType (self, node) :
        return self.snk.List(self.build(node.items))
    def build_TupleType (self, node) :
        return self.snk.Collection(self.snk.Instance(tuple),
                                   (self.build(node.items)))
    def build_SetType (self, node) :
        return self.snk.Set(self.build(node.items))
    def build_DictType (self, node) :
        return self.snk.Mapping(keys=self.build(node.keys),
                                items=self.build(node.items),
                                _dict=self.snk.Instance(self.snk.hdict))
    def build_EnumType (self, node) :
        return self.snk.OneOf(*(self._eval(child) for child in node.items))
    def build_NamedType (self, node) :
        name = node.name
        if name in self and self[name].kind == Decl.TYPE :
            return self[name].type
        elif name in self.globals :
            obj = self.globals[name]
            if inspect.isclass(obj) :
                return self.snk.Instance(obj)
            elif inspect.isroutine(obj) :
                return self.snk.TypeCheck(obj)
        elif hasattr(sys.modules["__builtin__"], name) :
            obj = getattr(sys.modules["__builtin__"], name)
            if inspect.isclass(obj) :
                return self.snk.Instance(obj)
            elif inspect.isroutine(obj) :
                return self.snk.TypeCheck(obj)
        self._raise(CompilationError,
                    "invalid type %r" % name)

if __name__ == "__main__" :
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE
                    | doctest.REPORT_ONLY_FIRST_FAILURE
                    | doctest.ELLIPSIS)
    from snakes.lang.abcd.parser import parse
    node = parse(open(sys.argv[1]).read())
    import snakes.plugins
    snk = snakes.plugins.load(["ops", "gv", "labels"], "snakes.nets", "snk")
    build = Builder(snk)
    net = build.build(node)
    net.draw(sys.argv[1] + ".png")



"""This package features the ABCD compiler, this is mainly a
command-line tool but it can be called also from Python. The API is
very simple and mimics the command line interface

### Function `snakes.utils.abcd.main.main` ###

    :::python
    def main (args=sys.argv[1:], src=None) : ...

Entry point of the compiler

##### Call API #####

  * `list args`:
  * `str src`:
  * `return PetriNet`:

##### Exceptions #####

  * `DeclarationError`: when
  * `CompilationError`: when

"""

# apidoc stop
from snakes import SnakesError

class CompilationError (SnakesError) :
    pass

class DeclarationError (SnakesError) :
    pass



"""
>>> testparser(Translator)
"""

import operator, sys
import snakes
from snakes.lang.python.parser import (ParseTree, ParseTestParser,
                                       Translator as PyTranslator,
                                       ParseTree as PyParseTree,
                                       testparser)
from snakes.lang.pgen import ParseError
from snakes.lang.ctlstar.pgen import parser
import snakes.lang.ctlstar.asdl as ast

_symbols = parser.tokenizer.tok_name.copy()
# next statement overrides 'NT_OFFSET' entry with 'single_input'
# (this is desired)
_symbols.update(parser.symbolMap)

def skip (token) :
    if token.kind == token.lexer.COMMENT :
        words = token.strip().split()
        if words[:2] == ["#", "coding="] :
            snakes.defaultencoding = words[2]
        elif words[:3] == ["#", "-*-", "coding:"] :
            snakes.defaultencoding = words[3]

parser.tokenizer.skip_token = skip

class ParseTree (PyParseTree) :
    _symbols = _symbols

class Translator (PyTranslator) :
    ParseTree = ParseTree
    parser = parser
    ST = ast
    def do_file_input (self, st, ctx) :
        """file_input: (NEWLINE | ctl_atomdef | ctl_propdef)* [ ctl_formula ] NEWLINE* ENDMARKER
        -> ast.Spec

        <<< atom foo () : return True
        ... prop bar () : True
        ... has(@'my place', x)
        "Spec(atoms=[Atom(name='foo', args=[], params=[], body=[Return(value=Name(id='True', ctx=Load()))])], properties=[Property(name='bar', args=[], params=[], body=Boolean(val=True))], main=InPlace(data=[Name(id='x', ctx=Load())], place=Place(name=None, place='my place')))"
        """
        atoms, props, main = [], [], None
        for i, child in enumerate(st) :
            if child.symbol == "ctl_atomdef" :
                atoms.append(self.do(child, ctx))
            elif child.symbol == "ctl_propdef" :
                props.append(self.do(child, ctx))
            elif child.symbol == "ctl_formula" :
                main = self.do(child, ctx)
            elif child.symbol in ("NEWLINE", "ENDMARKER") :
                pass
            else :
                raise ParseError(child.text, reason="unexpected token")
        return self.ST.Spec(lineno=st.srow, col_offset=st.scol,
                            atoms=atoms, properties=props, main=main)
    def do_ctl_atomdef (self, st, ctx) :
        """ctl_atomdef: 'atom' NAME '(' [ctl_parameters] ')' ':' suite
        -> ast.Atom

        <<< atom foo () : return True
        "Spec(atoms=[Atom(name='foo', args=[], params=[], body=[Return(value=Name(id='True', ctx=Load()))])], properties=[], main=None)"
        <<< atom bar () :
        ...     return True
        "Spec(atoms=[Atom(name='bar', args=[], params=[], body=[Return(value=Name(id='True', ctx=Load()))])], properties=[], main=None)"
        <<< atom egg (p = @'my place', x : int, q : place) :
        ...     return x in p and x in q
        "Spec(atoms=[Atom(name='egg', args=[Place(name='p', place='my place')], params=[Parameter(name='x', type='int'), Parameter(name='q', type='place')], body=[Return(value=BoolOp(op=And(), values=[Compare(left=Name(id='x', ctx=Load()), ops=[In()], comparators=[Name(id='p', ctx=Load())]), Compare(left=Name(id='x', ctx=Load()), ops=[In()], comparators=[Name(id='q', ctx=Load())])]))])], properties=[], main=None)"
        """
        if len(st) == 7 :
            args, params = self.do(st[3], ctx)
            return self.ST.Atom(lineno=st.srow, col_offset=st.scol,
                                name=st[1].text,
                                args=args,
                                params=params,
                                body=self.do(st[-1], ctx))
        else :
            return self.ST.Atom(lineno=st.srow, col_offset=st.scol,
                                name=st[1].text,
                                args=[],
                                params=[],
                                body=self.do(st[-1], ctx))
    def do_ctl_propdef (self, st, ctx) :
        """ctl_propdef: 'prop' NAME '(' [ctl_parameters] ')' ':' ctl_suite
        -> ast.Property

        <<< prop foo () : True
        "Spec(atoms=[], properties=[Property(name='foo', args=[], params=[], body=Boolean(val=True))], main=None)"
        <<< prop bar (p = @'my place', x : int, q : place) : True
        "Spec(atoms=[], properties=[Property(name='bar', args=[Place(name='p', place='my place')], params=[Parameter(name='x', type='int'), Parameter(name='q', type='place')], body=Boolean(val=True))], main=None)"
        <<< prop egg (p = @'my place', x : int, q : place) : True
        "Spec(atoms=[], properties=[Property(name='egg', args=[Place(name='p', place='my place')], params=[Parameter(name='x', type='int'), Parameter(name='q', type='place')], body=Boolean(val=True))], main=None)"
        """
        if len(st) == 7 :
            args, params = self.do(st[3], ctx)
            return self.ST.Property(lineno=st.srow, col_offset=st.scol,
                                    name=st[1].text,
                                    args=args,
                                    params=params,
                                    body=self.do(st[-1], ctx))
        else :
            return self.ST.Property(lineno=st.srow, col_offset=st.scol,
                                    name=st[1].text,
                                    args=[],
                                    params=[],
                                    body=self.do(st[-1], ctx))
    def do_ctl_suite (self, st, ctx) :
        """ctl_suite: ( ctl_formula NEWLINE | NEWLINE INDENT ctl_formula DEDENT )
        -> ast.form

        <<< prop foo () : True
        "Spec(atoms=[], properties=[Property(name='foo', args=[], params=[], body=Boolean(val=True))], main=None)"
        <<< prop bar () :
        ...    True
        "Spec(atoms=[], properties=[Property(name='bar', args=[], params=[], body=Boolean(val=True))], main=None)"
        """
        if len(st) == 2 :
            return self.do(st[0], ctx)
        else :
            return self.do(st[2], ctx)
    def do_ctl_parameters (self, st, ctx) :
        """ctl_parameters: (ctl_param ',')* ctl_param
        -> [ast.ctlarg], [ast.ctlparam]

        <<< prop foo (p = @'my place', q : place, x : int, r : place) : True
        "Spec(atoms=[], properties=[Property(name='foo', args=[Place(name='p', place='my place')], params=[Parameter(name='q', type='place'), Parameter(name='x', type='int'), Parameter(name='r', type='place')], body=Boolean(val=True))], main=None)"
        <<< prop bar (x : int) : True
        "Spec(atoms=[], properties=[Property(name='bar', args=[], params=[Parameter(name='x', type='int')], body=Boolean(val=True))], main=None)"
        <<< prop egg (p = @'my place') : True
        "Spec(atoms=[], properties=[Property(name='egg', args=[Place(name='p', place='my place')], params=[], body=Boolean(val=True))], main=None)"
        <<< prop spam (p = @'my place', q : int, p : place) : True
        Traceback (most recent call last):
        ...
        ParseError: ... duplicate parameter 'p'
        """
        args, params = [], []
        seen = set()
        for child in st[::2] :
            node = self.do(child, ctx)
            if node.name in seen :
                raise ParseError(child, reason="duplicate parameter %r"
                                 % node.name)
            seen.add(node.name)
            if isinstance(node, self.ST.Place) :
                args.append(node)
            else :
                params.append(node)
        return args, params
    def do_ctl_param (self, st, ctx) :
        """ctl_param: NAME ( '=' '@' STRING+ | ':' NAME )
        -> ast.ctlarg|ast.ctlparam

        <<< prop foo (p = @'my place', q : place, x : int, r : place) : True
        "Spec(atoms=[], properties=[Property(name='foo', args=[Place(name='p', place='my place')], params=[Parameter(name='q', type='place'), Parameter(name='x', type='int'), Parameter(name='r', type='place')], body=Boolean(val=True))], main=None)"
        <<< prop bar (x : int) : True
        "Spec(atoms=[], properties=[Property(name='bar', args=[], params=[Parameter(name='x', type='int')], body=Boolean(val=True))], main=None)"
        <<< prop egg (p = @'my place') : True
        "Spec(atoms=[], properties=[Property(name='egg', args=[Place(name='p', place='my place')], params=[], body=Boolean(val=True))], main=None)"
        """
        if st[1].text == "=" :
            return self.ST.Place(lineno=st.srow, col_offset=st.scol,
                                 name=st[0].text,
                                 place="".join(self.ST.literal_eval(c.text)
                                               for c in st[3:]))
        else :
            return self.ST.Parameter(lineno=st.srow, col_offset=st.scol,
                                     name=st[0].text,
                                     type=st[2].text)
    def do_ctl_arguments (self, st, ctx) :
        """ctl_arguments: (NAME '=' ctl_place_or_test ',')* NAME '=' ctl_place_or_test
        -> [(str, ast.expr)]

        <<< foo(x=3, p='my place')
        "Spec(atoms=[], properties=[], main=Instance(name='foo', args=[arg(arg='x', annotation=Num(n=3)), arg(arg='p', annotation=Str(s='my place'))]))"
        <<< foo(x=3)
        "Spec(atoms=[], properties=[], main=Instance(name='foo', args=[arg(arg='x', annotation=Num(n=3))]))"
        <<< foo(p='my place')
        "Spec(atoms=[], properties=[], main=Instance(name='foo', args=[arg(arg='p', annotation=Str(s='my place'))]))"
        <<< foo(x=3, p=@'my place')
        "Spec(atoms=[], properties=[], main=Instance(name='foo', args=[arg(arg='x', annotation=Num(n=3)), arg(arg='p', annotation=Place(name=None, place='my place'))]))"
        <<< foo(x=3)
        "Spec(atoms=[], properties=[], main=Instance(name='foo', args=[arg(arg='x', annotation=Num(n=3))]))"
        <<< foo(p=@'my place')
        "Spec(atoms=[], properties=[], main=Instance(name='foo', args=[arg(arg='p', annotation=Place(name=None, place='my place'))]))"
        """
        return [self.ST.arg(name.text, self.do(value, ctx))
                for name, value in zip(st[::4], st[2::4])]
    def do_ctl_place_or_test (self, st, ctx) :
        """ctl_place_or_test: test | '@' STRING+
        -> ast.expr | ast.Place

        <<< Foo(s='string', p=@'place', q=place_also)
        "Spec(atoms=[], properties=[], main=Instance(name='Foo', args=[arg(arg='s', annotation=Str(s='string')), arg(arg='p', annotation=Place(name=None, place='place')), arg(arg='q', annotation=Name(id='place_also', ctx=Load()))]))"
        """
        if st[0].symbol == "test" :
            return self.do(st[0], ctx)
        else :
            return self.do_ctl_place(st, ctx)
    def do_ctl_formula (self, st, ctx) :
        """ctl_formula: ctl_or_formula [ ctl_connector ctl_or_formula ]
        -> ast.form

        <<< True
        'Spec(atoms=[], properties=[], main=Boolean(val=True))'
        <<< False => True
        'Spec(atoms=[], properties=[], main=CtlBinary(op=Imply(), left=Boolean(val=False), right=Boolean(val=True)))'
        <<< False <=> False
        'Spec(atoms=[], properties=[], main=CtlBinary(op=Iff(), left=Boolean(val=False), right=Boolean(val=False)))'
        """
        if len(st) == 1 :
            return self.do(st[0], ctx)
        else :
            return self.ST.CtlBinary(lineno=st.srow, col_offset=st.scol,
                                     op=self.do_ctl_connector(st[1], ctx),
                                     left=self.do(st[0], ctx),
                                     right=self.do(st[2], ctx))
    def do_ctl_connector (self, st, ctx) :
        """ctl_connector: ( '=' '>' | '<=' '>' )
        -> ast.ctlbinary

        <<< False => True
        'Spec(atoms=[], properties=[], main=CtlBinary(op=Imply(), left=Boolean(val=False), right=Boolean(val=True)))'
        <<< False <=> False
        'Spec(atoms=[], properties=[], main=CtlBinary(op=Iff(), left=Boolean(val=False), right=Boolean(val=False)))'
        """
        op = "".join(child.text for child in st)
        return self._ctl_binary_op[op](lineno=st.srow,
                                       col_offset=st.scol)

    def do_ctl_or_formula (self, st, ctx) :
        """ctl_or_formula: ctl_and_formula ('or' ctl_and_formula)*
        -> ast.form

        <<< True or False
        'Spec(atoms=[], properties=[], main=CtlBinary(op=Or(), left=Boolean(val=True), right=Boolean(val=False)))'
        <<< True or False or True
        'Spec(atoms=[], properties=[], main=CtlBinary(op=Or(), left=CtlBinary(op=Or(), left=Boolean(val=True), right=Boolean(val=False)), right=Boolean(val=True)))'
        <<< True or False or False and True and False
        'Spec(atoms=[], properties=[], main=CtlBinary(op=Or(), left=CtlBinary(op=Or(), left=Boolean(val=True), right=Boolean(val=False)), right=CtlBinary(op=And(), left=CtlBinary(op=And(), left=Boolean(val=False), right=Boolean(val=True)), right=Boolean(val=False))))'
        """
        if len(st) == 1 :
            return self.do(st[0], ctx)
        else :
            values = [self.do(child, ctx) for child in st[::2]]
            ops = [self._ctl_binary_op[child.text](lineno=child.srow,
                                                   col_offset=child.scol)
                   for child in st[1::2]]
            while len(values) > 1 :
                left = values.pop(0)
                right = values.pop(0)
                operator = ops.pop(0)
                values.insert(0, self.ST.CtlBinary(lineno=st.srow,
                                                   col_offset=st.scol,
                                                   left=left,
                                                   op=operator,
                                                   right=right))
            return values[0]
    def do_ctl_and_formula (self, st, ctx) :
        """ctl_and_formula: ctl_not_formula ('and' ctl_not_formula)*
        -> ast.form

        <<< True and False
        'Spec(atoms=[], properties=[], main=CtlBinary(op=And(), left=Boolean(val=True), right=Boolean(val=False)))'
        """
        return self.do_ctl_or_formula(st, ctx)
    def do_ctl_not_formula (self, st, ctx) :
        """ctl_not_formula: ('not' ctl_not_formula | ctl_binary_formula)
        -> ast.form

        <<< True
        'Spec(atoms=[], properties=[], main=Boolean(val=True))'
        <<< not True
        'Spec(atoms=[], properties=[], main=CtlUnary(op=Not(), child=Boolean(val=True)))'
        <<< not not True
        'Spec(atoms=[], properties=[], main=CtlUnary(op=Not(), child=CtlUnary(op=Not(), child=Boolean(val=True))))'
        """
        if len(st) == 1 :
            return self.do(st[0], ctx)
        else :
            return self.ST.CtlUnary(lineno=st.srow, col_offset=st.scol,
                                    op=self.ST.Not(lineno=st[0].srow,
                                                   col_offset=st[0].scol),
                                    child=self.do(st[1], ctx))
    def do_ctl_binary_formula (self, st, ctx) :
        """ctl_binary_formula: ctl_unary_formula [ ctl_binary_op ctl_unary_formula ]
        -> ast.form

        <<< True U False
        'Spec(atoms=[], properties=[], main=CtlBinary(op=Until(), left=Boolean(val=True), right=Boolean(val=False)))'
        <<< True W False
        'Spec(atoms=[], properties=[], main=CtlBinary(op=WeakUntil(), left=Boolean(val=True), right=Boolean(val=False)))'
        <<< True R False
        'Spec(atoms=[], properties=[], main=CtlBinary(op=Release(), left=Boolean(val=True), right=Boolean(val=False)))'
        """
        if len(st) == 1 :
            return self.do(st[0], ctx)
        else :
            return self.ST.CtlBinary(lineno=st.srow, col_offset=st.scol,
                                     op=self.do(st[1], ctx),
                                     left=self.do(st[0], ctx),
                                     right=self.do(st[2], ctx))
    def do_ctl_unary_formula (self, st, ctx) :
        """ctl_unary_formula: [ ctl_unary_op ] (ctl_atom_formula | '(' ctl_formula ')')
        -> ast.form

        <<< True
        'Spec(atoms=[], properties=[], main=Boolean(val=True))'
        <<< X True
        'Spec(atoms=[], properties=[], main=CtlUnary(op=Next(), child=Boolean(val=True)))'
        <<< A True
        'Spec(atoms=[], properties=[], main=CtlUnary(op=All(), child=Boolean(val=True)))'
        <<< G True
        'Spec(atoms=[], properties=[], main=CtlUnary(op=Globally(), child=Boolean(val=True)))'
        <<< F True
        'Spec(atoms=[], properties=[], main=CtlUnary(op=Future(), child=Boolean(val=True)))'
        <<< E True
        'Spec(atoms=[], properties=[], main=CtlUnary(op=Exists(), child=Boolean(val=True)))'
        <<< (True or False)
        'Spec(atoms=[], properties=[], main=CtlBinary(op=Or(), left=Boolean(val=True), right=Boolean(val=False)))'
        <<< X (True or False)
        'Spec(atoms=[], properties=[], main=CtlUnary(op=Next(), child=CtlBinary(op=Or(), left=Boolean(val=True), right=Boolean(val=False))))'
        """
        if len(st) == 1 :
            return self.do(st[0], ctx)
        elif len(st) == 2 :
            return self.ST.CtlUnary(lineno=st.srow, col_offset=st.scol,
                                    op=self.do(st[0], ctx),
                                    child=self.do(st[1], ctx))
        elif len(st) == 3 :
            return self.do(st[1], ctx)
        else :
            return self.ST.CtlUnary(lineno=st.srow, col_offset=st.scol,
                                    op=self.do(st[0], ctx),
                                    child=self.do(st[2], ctx))
    _ctl_unary_op = {"A" : ast.All,
                     "E" : ast.Exists,
                     "X" : ast.Next,
                     "F" : ast.Future,
                     "G" : ast.Globally}
    def do_ctl_unary_op (self, st, ctx) :
        """ctl_unary_op: ('A' | 'G' | 'F' | 'E' | 'X')
        -> ast.ctlunary

        <<< X True
        'Spec(atoms=[], properties=[], main=CtlUnary(op=Next(), child=Boolean(val=True)))'
        <<< A True
        'Spec(atoms=[], properties=[], main=CtlUnary(op=All(), child=Boolean(val=True)))'
        <<< G True
        'Spec(atoms=[], properties=[], main=CtlUnary(op=Globally(), child=Boolean(val=True)))'
        <<< F True
        'Spec(atoms=[], properties=[], main=CtlUnary(op=Future(), child=Boolean(val=True)))'
        <<< E True
        'Spec(atoms=[], properties=[], main=CtlUnary(op=Exists(), child=Boolean(val=True)))'
        """
        return self._ctl_unary_op[st[0].text](lineno=st.srow,
                                              col_offset=st.scol)
    _ctl_binary_op = {"=>"  : ast.Imply,
                      "<=>" : ast.Iff,
                      "and" : ast.And,
                      "or"  : ast.Or,
                      "U"   : ast.Until,
                      "W"   : ast.WeakUntil,
                      "R"   : ast.Release}
    def do_ctl_binary_op (self, st, ctx) :
        """ctl_binary_op: ('R' | 'U' | 'W')
        -> ast.ctlbinary

        <<< True R False
        'Spec(atoms=[], properties=[], main=CtlBinary(op=Release(), left=Boolean(val=True), right=Boolean(val=False)))'
        <<< True U False
        'Spec(atoms=[], properties=[], main=CtlBinary(op=Until(), left=Boolean(val=True), right=Boolean(val=False)))'
        <<< True W False
        'Spec(atoms=[], properties=[], main=CtlBinary(op=WeakUntil(), left=Boolean(val=True), right=Boolean(val=False)))'
        """
        return self._ctl_binary_op[st[0].text](lineno=st[0].srow,
                                               col_offset=st[0].scol)
    def do_ctl_atom_formula (self, st, ctx) :
        """ctl_atom_formula: ( 'empty' '(' ctl_place ')'
                   | 'marked' '(' ctl_place ')'
		   | 'has' ['not'] '(' ctl_place ',' test (',' test)* ')'
		   | 'deadlock' | 'True' | 'False'
		   | NAME '(' ctl_arguments ')'
		   | 'forall' [ 'distinct' ] NAME (',' NAME)*
                     'in' ctl_place '(' ctl_atom_formula ')'
		   | 'exists' [ 'distinct' ] NAME (',' NAME)*
                     'in' ctl_place '(' ctl_atom_formula ')' )
        -> ast.atom

        <<< empty(p)
        "Spec(atoms=[], properties=[], main=EmptyPlace(place=Parameter(name='p', type='place')))"
        <<< empty(@'my' 'place')
        "Spec(atoms=[], properties=[], main=EmptyPlace(place=Place(name=None, place='myplace')))"
        <<< marked(p)
        "Spec(atoms=[], properties=[], main=MarkedPlace(place=Parameter(name='p', type='place')))"
        <<< marked(@'my place')
        "Spec(atoms=[], properties=[], main=MarkedPlace(place=Place(name=None, place='my place')))"
        <<< has(p, x)
        "Spec(atoms=[], properties=[], main=InPlace(data=[Name(id='x', ctx=Load())], place=Parameter(name='p', type='place')))"
        <<< has not(p, x, y)
        "Spec(atoms=[], properties=[], main=NotInPlace(data=[Name(id='x', ctx=Load()), Name(id='y', ctx=Load())], place=Parameter(name='p', type='place')))"
        <<< deadlock
        'Spec(atoms=[], properties=[], main=Deadlock())'
        <<< True
        'Spec(atoms=[], properties=[], main=Boolean(val=True))'
        <<< False
        'Spec(atoms=[], properties=[], main=Boolean(val=False))'
        <<< myprop(x=1, p='my place')
        "Spec(atoms=[], properties=[], main=Instance(name='myprop', args=[arg(arg='x', annotation=Num(n=1)), arg(arg='p', annotation=Str(s='my place'))]))"
        <<< forall x in p (has(q, y))
        "Spec(atoms=[], properties=[], main=Quantifier(op=All(), vars=['x'], place=Parameter(name='p', type='place'), child=InPlace(data=[Name(id='y', ctx=Load())], place=Parameter(name='q', type='place')), distinct=False))"
        <<< forall x, y in p (has(q, x+y, x-y))
        "Spec(atoms=[], properties=[], main=Quantifier(op=All(), vars=['x', 'y'], place=Parameter(name='p', type='place'), child=InPlace(data=[BinOp(left=Name(id='x', ctx=Load()), op=Add(), right=Name(id='y', ctx=Load())), BinOp(left=Name(id='x', ctx=Load()), op=Sub(), right=Name(id='y', ctx=Load()))], place=Parameter(name='q', type='place')), distinct=False))"
        <<< forall distinct x, y in p (has(q, x+y, x-y))
        "Spec(atoms=[], properties=[], main=Quantifier(op=All(), vars=['x', 'y'], place=Parameter(name='p', type='place'), child=InPlace(data=[BinOp(left=Name(id='x', ctx=Load()), op=Add(), right=Name(id='y', ctx=Load())), BinOp(left=Name(id='x', ctx=Load()), op=Sub(), right=Name(id='y', ctx=Load()))], place=Parameter(name='q', type='place')), distinct=True))"
        <<< exists x in p (has(q, y))
        "Spec(atoms=[], properties=[], main=Quantifier(op=Exists(), vars=['x'], place=Parameter(name='p', type='place'), child=InPlace(data=[Name(id='y', ctx=Load())], place=Parameter(name='q', type='place')), distinct=False))"
        <<< exists x, y in p (has(q, x+y, x-y))
        "Spec(atoms=[], properties=[], main=Quantifier(op=Exists(), vars=['x', 'y'], place=Parameter(name='p', type='place'), child=InPlace(data=[BinOp(left=Name(id='x', ctx=Load()), op=Add(), right=Name(id='y', ctx=Load())), BinOp(left=Name(id='x', ctx=Load()), op=Sub(), right=Name(id='y', ctx=Load()))], place=Parameter(name='q', type='place')), distinct=False))"
        <<< exists distinct x, y in p (has(q, x+y, x-y))
        "Spec(atoms=[], properties=[], main=Quantifier(op=Exists(), vars=['x', 'y'], place=Parameter(name='p', type='place'), child=InPlace(data=[BinOp(left=Name(id='x', ctx=Load()), op=Add(), right=Name(id='y', ctx=Load())), BinOp(left=Name(id='x', ctx=Load()), op=Sub(), right=Name(id='y', ctx=Load()))], place=Parameter(name='q', type='place')), distinct=True))"
        """
        if st[0].text in ("True", "False") :
            return self.ST.Boolean(lineno=st.srow, col_offset=st.scol,
                                   val=(st[0].text  == "True"))
        elif st[0].text == "deadlock" :
            return self.ST.Deadlock(lineno=st.srow, col_offset=st.scol)
        elif st[0].text in ("empty", "marked") :
            node = (self.ST.EmptyPlace if st[0].text == "empty"
                    else self.ST.MarkedPlace)
            return node(lineno=st.srow, col_offset=st.scol,
                        place=self.do(st[2], ctx))
        elif st[0].text == "has" :
            if st[1].text == "not" :
                node = self.ST.NotInPlace
                start = 3
            else :
                node = self.ST.InPlace
                start = 2
            place = self.do(st[start], ctx)
            if (isinstance(place, self.ST.Parameter)
                and place.type != "place") :
                raise ParseError(st[-4], reason="'place' parameter expected")
            return node(lineno=st.srow, col_offset=st.scol,
                        data=[self.do(c, ctx) for c in st[start+2::2]],
                        place=place)
        elif st[0].text in ("forall", "exists") :
            op = (self.ST.All if st[0].text == "forall"
                  else self.ST.Exists)
            distinct = st[1].text == "distinct"
            start = 2 if distinct else 1
            return self.ST.Quantifier(lineno=st.srow, col_offset=st.scol,
                                      op=op(),
                                      vars=[c.text for c in st[start:-5:2]],
                                      place=self.do(st[-4], ctx),
                                      child=self.do(st[-2], ctx),
                                      distinct=distinct)
        else :
            return self.ST.Instance(lineno=st.srow, col_offset=st.scol,
                                    name=st[0].text,
                                    args=self.do(st[2], ctx))
    def do_ctl_place (self, st, ctx) :
        """ctl_place: '@' STRING+ | NAME
        -> ast.ctlarg

        <<< has(@'my place', x)
        "Spec(atoms=[], properties=[], main=InPlace(data=[Name(id='x', ctx=Load())], place=Place(name=None, place='my place')))"
        <<< has(@'another' 'place', y)
        "Spec(atoms=[], properties=[], main=InPlace(data=[Name(id='y', ctx=Load())], place=Place(name=None, place='anotherplace')))"
        <<< has(@'my place', x, y)
        "Spec(atoms=[], properties=[], main=InPlace(data=[Name(id='x', ctx=Load()), Name(id='y', ctx=Load())], place=Place(name=None, place='my place')))"
        <<< has not(@'my place', x)
        "Spec(atoms=[], properties=[], main=NotInPlace(data=[Name(id='x', ctx=Load())], place=Place(name=None, place='my place')))"
        <<< has not(@'my place', x, y)
        "Spec(atoms=[], properties=[], main=NotInPlace(data=[Name(id='x', ctx=Load()), Name(id='y', ctx=Load())], place=Place(name=None, place='my place')))"
        """
        if st[0].symbol == "NAME" :
            return self.ST.Parameter(lineno=st.srow, col_offset=st.scol,
                                     name=st[0].text,
                                     type="place")
        else :
            return self.ST.Place(lineno=st.srow, col_offset=st.scol,
                                 name=None,
                                 place="".join(self.ST.literal_eval(c.text)
                                               for c in st[1:]))

parse = Translator.parse

if __name__ == "__main__" :
    testparser(Translator)



"""This plugin defines a data structure `Cluster` that allows to group
nodes hierarchically. This is used by plugin `gv` to improve the
graphical layout of Petri nets obtained using plugin `ops`.

In general, this plugin is probably not needed by anyone,
consequently, documentation will be very terse.
"""

import snakes.plugins
from snakes.plugins import new_instance
from snakes.pnml import Tree
from snakes.data import iterate

class Cluster (object) :
    """A hierarchical data structure to organise strings (intended to
    be node names).
    """
    def __init__ (self, nodes=[], children=[]) :
        """Create a cluster whose to-level nodes are `nodes` and with
        sub-clusters given in `children`.

        >>> Cluster(['a', 'b'],
        ...         [Cluster(['1', '2'],
        ...                  [Cluster(['A'])]),
        ...          Cluster(['3', '4', '5'],
        ...                  [Cluster(['C', 'D'])])])
        Cluster(...)
        """
        self._nodes = set(nodes)
        self._children = []
        self._cluster = {}
        for child in children :
            self.add_child(child)
    __pnmltag__ = "clusters"
    # apidoc skip
    def __pnmldump__ (self) :
        """
        >>> Cluster(['a', 'b'],
        ...         [Cluster(['1', '2'],
        ...                  [Cluster(['A'])]),
        ...          Cluster(['3', '4', '5'],
        ...                  [Cluster(['C', 'D'])])]).__pnmldump__()
        <?xml version="1.0" encoding="utf-8"?>
        <pnml>...
         <clusters>
          <node>
          ...
          </node>
          <node>
          ...
          </node>
          <clusters>
          ...
          </clusters>
         </clusters>
        </pnml>
        """
        result = Tree(self.__pnmltag__, None)
        for node in self._nodes :
            result.add_child(Tree("node", node))
        for child in self._children :
            result.add_child(Tree.from_obj(child))
        return result
    # apidoc skip
    @classmethod
    def __pnmlload__ (cls, tree) :
        """
        >>> t = Cluster(['a', 'b'],
        ...             [Cluster(['1', '2'],
        ...                      [Cluster(['A'])]),
        ...              Cluster(['3', '4', '5'],
        ...                      [Cluster(['C', 'D'])])]).__pnmldump__()
        >>> Cluster.__pnmlload__(t)
        Cluster(['...', '...'],
                [Cluster(['...', '...'],
                         [Cluster(['A'], [])]),
                 Cluster(['...', '...', '...'],
                         [Cluster(['...', '...'], [])])])
        """

        result = cls()
        for child in tree.children :
            if child.name == "node" :
                result.add_node(child.data)
            else :
                result.add_child(child.to_obj())
        return result
    # apidoc skip
    def __str__ (self) :
        return "cluster_%s" % str(id(self)).replace("-", "m")
    # apidoc skip
    def __repr__ (self) :
        """
        >>> Cluster(['a', 'b'],
        ...         [Cluster(['1', '2'],
        ...                  [Cluster(['A'])]),
        ...          Cluster(['3', '4', '5'],
        ...                  [Cluster(['C', 'D'])])])
        Cluster(['...', '...'],
                [Cluster(['...', '...'],
                         [Cluster(['A'], [])]),
                 Cluster(['...', '...', '...'],
                         [Cluster(['...', '...'], [])])])
        """
        return "%s([%s], [%s])" % (self.__class__.__name__,
                                   ", ".join(repr(n) for n in self.nodes()),
                                   ", ".join(repr(c) for c in self.children()))
    # apidoc skip
    def copy (self) :
        """
        >>> Cluster(['a', 'b'],
        ...         [Cluster(['1', '2'],
        ...                  [Cluster(['A'])]),
        ...          Cluster(['3', '4', '5'],
        ...                  [Cluster(['C', 'D'])])]).copy()
        Cluster(['...', '...'],
                [Cluster(['...', '...'],
                         [Cluster(['A'], [])]),
                 Cluster(['...', '...', '...'],
                         [Cluster(['...', '...'], [])])])
        """
        return self.__class__(self._nodes,
                              (child.copy() for child in self._children))
    def get_path (self, name) :
        """Get the path of a name inside the cluster. This path is a
        list of indexes for each child-cluster within its parent.

        >>> Cluster(['a', 'b'],
        ...         [Cluster(['1', '2'],
        ...                  [Cluster(['A'])]),
        ...          Cluster(['3', '4', '5'],
        ...                  [Cluster(['C', 'D'])])]).get_path('C')
        [1, 0]

        @param name: the searched name
        @type name: `str`
        @return: the list of indexes as `int` values
        @rtype: `list`
        """
        if name in self._nodes :
            return []
        else :
            for num, child in enumerate(self._children) :
                if name in child :
                    return [num] + child.get_path(name)
    def add_node (self, name, path=None) :
        """Add `name` to the cluster, optionally at a given position
        `path`.

        >>> c = Cluster(['a', 'b'],
        ...             [Cluster(['1', '2'],
        ...                      [Cluster(['A'])]),
        ...              Cluster(['3', '4', '5'],
        ...                      [Cluster(['C', 'D'])])])
        >>> c.add_node('c')
        Cluster([...'c'...], ...)
        >>> c.add_node('E', [1, 0])
        Cluster([...'E'...], [])
        >>> c
        Cluster([...'c'...],
                [Cluster(['...', '...'],
                         [Cluster(['A'], [])]),
                 Cluster(['...', '...', '...'],
                         [Cluster([...'E'...], [])])])

        @param name: name to add
        @type name: `str`
        @param path: position where `name`should be added, given as a
            list of indexes
        @type path: `list`
        """
        if path in (None, [], ()) :
            self._nodes.add(name)
            return self
        else :
            while len(self._children) <= path[0] :
                self._children.append(self.__class__())
            target = self._children[path[0]].add_node(name, path[1:])
            self._cluster[name] = target
            return target
    def remove_node (self, name) :
        """Remove a name from the cluster.

        >>> c = Cluster(['a', 'b'],
        ...             [Cluster(['1', '2'],
        ...                      [Cluster(['A'])]),
        ...              Cluster(['3', '4', '5'],
        ...                      [Cluster(['C', 'D'])])])
        >>> c.remove_node('4')
        >>> c
        Cluster(['...', '...'],
                [Cluster(['...', '...'],
                         [Cluster(['A'], [])]),
                 Cluster(['...', '...'],
                         [Cluster(['...', '...'], [])])])

        @param name: name to remove
        @type name: `str`
        """
        if name in self._cluster :
            self._cluster[name].remove_node(name)
        else :
            self._nodes.remove(name)
    def rename_node (self, old, new) :
        """Change a name in the cluster.

        >>> c = Cluster(['a', 'b'],
        ...             [Cluster(['1', '2'],
        ...                      [Cluster(['A'])]),
        ...              Cluster(['3', '4', '5'],
        ...                      [Cluster(['C', 'D'])])])
        >>> c.rename_node('4', '42')
        >>> c
        Cluster(['...', '...'],
                [Cluster(['...', '...'],
                         [Cluster(['A'], [])]),
                 Cluster([...'42'...],
                         [Cluster(['...', '...'], [])])])

        @param old: name to change
        @type old: `str`
        @param new: new name to replace `old`
        @type new: `str`
        """
        if old in self._cluster :
            self._cluster[old].rename_node(old, new)
            self._cluster[new] = self._cluster[old]
            del self._cluster[old]
        elif old in self._nodes :
            self._nodes.remove(old)
            self._nodes.add(new)
        else :
            for child in self.children() :
                child.rename_node(old, new)
    def add_child (self, cluster=None) :
        """Add a child cluster

        >>> c = Cluster(['a', 'b'],
        ...             [Cluster(['1', '2'],
        ...                      [Cluster(['A'])]),
        ...              Cluster(['3', '4', '5'],
        ...                      [Cluster(['C', 'D'])])])
        >>> c.add_child(c.copy())
        >>> c
        Cluster(['...', '...'],
                [Cluster(['...', '...'],
                         [Cluster(['A'], [])]),
                 Cluster(['...', '...', '...'],
                         [Cluster(['...', '...'], [])]),
                 Cluster(['...', '...'],
                         [Cluster(['...', '...'],
                                  [Cluster(['A'], [])]),
                          Cluster(['...', '...', '...'],
                                  [Cluster(['...', '...'], [])])])])

        @param cluster: the new child, if `None` is given, an empty
            child is added
        @type cluster: `Cluster`
        """
        if cluster is None :
            cluster = Cluster()
        self._cluster.update(cluster._cluster)
        for node in cluster._nodes :
            self._cluster[node] = cluster
        self._children.append(cluster)
    def nodes (self, all=False) :
        """Returns the nodes in the cluster: only the top-level ones
        is `all` is `False`, or all the nodes otherwise.

        >>> list(sorted(Cluster(['a', 'b'],
        ...         [Cluster(['1', '2'],
        ...                  [Cluster(['A'])]),
        ...          Cluster(['3', '4', '5'],
        ...                  [Cluster(['C', 'D'])])]).nodes()))
        ['a', 'b']
        >>> list(sorted(Cluster(['a', 'b'],
        ...         [Cluster(['1', '2'],
        ...                  [Cluster(['A'])]),
        ...          Cluster(['3', '4', '5'],
        ...                  [Cluster(['C', 'D'])])]).nodes(True)))
        ['1', '2', '3', '4', '5', 'A', 'C', 'D', 'a', 'b']

        @param all: whether all the nodes should be returned or only
            the top-level ones
        @type all: `bool`
        @return: list of nodes
        @rtype: `list`
        """
        if all :
            result = set()
            for cluster in self :
                result.update(cluster.nodes())
            return result
        else :
            return set(self._nodes)
    def children (self) :
        """Return the children of the cluster.

        >>> Cluster(['a', 'b'],
        ...         [Cluster(['1', '2'],
        ...                  [Cluster(['A'])]),
        ...          Cluster(['3', '4', '5'],
        ...                  [Cluster(['C', 'D'])])]).children()
        (Cluster(['...', '...'],
                 [Cluster(['A'], [])]),
         Cluster(['...', '...', '...'],
                 [Cluster(['...', '...'], [])]))

        @return: the children of `self`
        @rtype: `tuple`
        """
        return tuple(self._children)
    def __contains__ (self, name) :
        """Test if a name is in the cluster.

        >>> c = Cluster(['a', 'b'],
        ...             [Cluster(['1', '2'],
        ...                      [Cluster(['A'])]),
        ...              Cluster(['3', '4', '5'],
        ...                      [Cluster(['C', 'D'])])])
        >>> 'a' in c
        True
        >>> 'x' in c
        False
        >>> '4' in c
        True

        @param name: the node to test
        @type name: `str`
        @return: whether `name` is in the cluster
        @rtype: `bool`
        """
        if name in self._nodes :
            return True
        for child in self._children :
            if name in child :
                return True
        return False
    def __iter__ (self) :
        """Iterate over the clusters and its children, yielding lists
        of nodes at each level.

        >>> c = Cluster(['a', 'b'],
        ...             [Cluster(['1', '2'],
        ...                      [Cluster(['A'])]),
        ...              Cluster(['3', '4', '5'],
        ...                      [Cluster(['C', 'D'])])])
        >>> for cluster in c :
        ...    print(list(sorted(cluster.nodes())))
        ['a', 'b']
        ['1', '2']
        ['A']
        ['3', '4', '5']
        ['C', 'D']
        """
        yield self
        for child in self._children :
            for item in child :
                yield item

@snakes.plugins.plugin("snakes.nets")
def extend (module) :
    class PetriNet (module.PetriNet) :
        """Class `PetriNet`is extended so that instances have an
        attribute `clusters` to which the nodes are added.
        """
        def add_place (self, place, **options) :
            """
            @param place: the place to add
            @type place: `Place`
            @param options: additional options for plugins
            @keyword cluster: position of the new place in the cluster
            """
            path = options.pop("cluster", None)
            module.PetriNet.add_place(self, place, **options)
            self.clusters.add_node(place.name, path)
        def add_transition (self, trans, **options) :
            """
            @param trans: the transition to add
            @type trans: `Transition`
            @param options: additional options for plugins
            @keyword cluster: position of the new transition in the
                cluster
            """
            path = options.pop("cluster", None)
            module.PetriNet.add_transition(self, trans, **options)
            self.clusters.add_node(trans.name, path)
        # apidoc stop
        def __init__ (self, name, **options) :
            module.PetriNet.__init__(self, name, **options)
            self.clusters = Cluster()
        def copy (self, name=None, **options) :
            result = module.PetriNet.copy(self, name, **options)
            result.clusters = self.clusters.copy()
            return result
        def __pnmldump__ (self) :
            result = module.PetriNet.__pnmldump__(self)
            result.add_child(Tree.from_obj(self.clusters))
            return result
        @classmethod
        def __pnmlload__ (cls, tree) :
            result = new_instance(cls, module.PetriNet.__pnmlload__(tree))
            result.clusters = tree.child(Cluster.__pnmltag__).to_obj()
            return result
        def remove_place (self, name, **options) :
            module.PetriNet.remove_place(self, name, **options)
            self.clusters.remove_node(name)
        def remove_transition (self, name, **options) :
            module.PetriNet.remove_transition(self, name, **options)
            self.clusters.remove_node(name)
        def rename_node (self, old, new, **options) :
            module.PetriNet.rename_node(self, old, new, **options)
            self.clusters.rename_node(old, new)
    return PetriNet, Cluster



"""A plugin to implement boundedness on places.

When this plugin is loaded, `Place` constructor accepts a parameter
`bound` that allows to specify the minimal and maximal number of
tokens that the place is allowed to carry. (Repeated tokens count as
many times as they are repeated.) Exception `ConstraintError` is
raised whenever an operation (like a transition firing) leads to
violate any of a place bound. (But note that direct modifications of a
`Place.tokens` are not checked.)

>>> import snakes.plugins
>>> snakes.plugins.load('bound', 'snakes.nets', 'nets')
<module ...>
>>> from nets import *

If parameter `bound` is given as a non-negative integer, this is the
upper bound of the place (and its lower bound is zero).

>>> n = PetriNet('N')
>>> p = Place('p', [dot], bound=3)
>>> n.add_place(p)
>>> put = Transition('put')
>>> n.add_transition(put)
>>> n.add_output('p', 'put', Value(dot))
>>> p.tokens
MultiSet([dot])
>>> put.fire(Substitution())
>>> p.tokens
MultiSet([dot, dot])
>>> put.fire(Substitution())
>>> p.tokens
MultiSet([dot, dot, dot])
>>> put.fire(Substitution())
Traceback (most recent call last):
  ...
ConstraintError: upper bound of place 'p' reached

If `bound` is given as a pair `(n, None)` where `n` is a non-negative
integer, then `n` is the lower bound of the place (and it has no upper
bound).

>>> n = PetriNet('N')
>>> p = Place('p', [dot, dot], bound=(1, None))
>>> n.add_place(p)
>>> put = Transition('put')
>>> n.add_transition(put)
>>> n.add_output('p', 'put', Value(dot))
>>> get =  Transition('get')
>>> n.add_transition(get)
>>> n.add_input('p', 'get', Value(dot))
>>> p.tokens
MultiSet([dot, dot])
>>> get.fire(Substitution())
>>> p.tokens
MultiSet([dot])
>>> get.fire(Substitution())
Traceback (most recent call last):
  ...
ConstraintError: lower bound of place 'p' reached
>>> for i in range(100) : # no upper bound
...     put.fire(Substitution())

If `bound` is given as a pair of non-negative integers `(n,m)` such
that `n <= m` then `n` is the lower bound of the place and `m` its
upper bound.

>>> n = PetriNet('N')
>>> p = Place('p', [dot, dot], bound=(1, 3))
>>> n.add_place(p)
>>> put = Transition('put')
>>> n.add_transition(put)
>>> n.add_output('p', 'put', Value(dot))
>>> get =  Transition('get')
>>> n.add_transition(get)
>>> n.add_input('p', 'get', Value(dot))
>>> p.tokens
MultiSet([dot, dot])
>>> put.fire(Substitution())
>>> p.tokens
MultiSet([dot, dot, dot])
>>> put.fire(Substitution())
Traceback (most recent call last):
  ...
ConstraintError: upper bound of place 'p' reached
>>> get.fire(Substitution())
>>> p.tokens
MultiSet([dot, dot])
>>> get.fire(Substitution())
>>> p.tokens
MultiSet([dot])
>>> get.fire(Substitution())
Traceback (most recent call last):
  ...
ConstraintError: lower bound of place 'p' reached

Any other value for bound is refused raising a `ValueError`.

"""

from snakes.data import iterate
from snakes import ConstraintError
import snakes.plugins

@snakes.plugins.plugin("snakes.nets")
def extend (module) :
    "Extends `module`"
    class Place (module.Place) :
        "Extend places with boundedness"
        def __init__ (self, name, tokens, check=None, **args) :
            """Add new keyword argument `bound`

            @param args: plugin options
            @keyword bound: the place boundaries
            @type bound: `int` or `(int, None)` or `(int, int)`
            """
            bound = args.pop("bound", (0, None))
            if isinstance(bound, int) and bound >= 0 :
                self._bound_min, self._bound_max = 0, bound
            elif (isinstance(bound, tuple) and len(bound) == 2
                  and isinstance(bound[0], int) and bound[0] >= 0
                  and ((isinstance(bound[1], int) and bound[1] >= bound[0])
                       or bound[1] is None)) :
                self._bound_min, self._bound_max = bound
            else :
                raise ValueError("invalid value for parameter 'bound'")
            module.Place.__init__(self, name, tokens, check, **args)
        def add (self, tokens) :
            """Add tokens to the place.

            @param tokens: a collection of tokens to be added to the
                place, note that `str` are not considered as iterable and
                used a a single value instead of as a collection
            @type tokens: `collection`
            """
            if (self._bound_max is not None
                and len(self.tokens) + len(list(iterate(tokens))) > self._bound_max) :
                raise ConstraintError("upper bound of place %r reached" % self.name)
            module.Place.add(self, tokens)
        def remove (self, tokens) :
            """Remove tokens from the place.

            @param tokens: a collection of tokens to be removed from the
                place, note that `str` are not considered as iterable and
                used a a single value instead of as a collection
            @type tokens: `collection`
            """
            if len(self.tokens) - len(list(iterate(tokens))) < self._bound_min :
                raise ConstraintError("lower bound of place %r reached" % self.name)
            module.Place.remove(self, tokens)
        def reset (self, tokens) :
            """Replace the marking with `tokens`.

            @param tokens: a collection of tokens to be removed from the
                place, note that `str` are not considered as iterable and
                used a a single value instead of as a collection
            @type tokens: `collection`
            """
            count = len(list(iterate(tokens)))
            bmax = count if self._bound_max is None else self._bound_max
            if not (self._bound_min <= count <= bmax) :
                raise ConstraintError("not within bounds of place %r" % self.name)
            module.Place.reset(self, tokens)
    return Place



import sys, datetime
from snakes.lang.pylib import asdl
from collections import defaultdict
from functools import partial

class memoize(object):
    def __init__(self, function):
        self.function = function
        self.memoized = {}

    def __call__(self, *args):
        try:
            return self.memoized[args]
        except KeyError:
            call = self.function(*args)
            self.memoized[args] = call
            return call

    def __get__(self, obj, objtype):
        """Support instance methods.
        """
        return partial(self.__call__, obj)

def component_has_cycle(node, graph, proceeding, visited):
    if node in visited:
        return False
    if node in proceeding:
        proceeding.append(node) # populate trace
        return True
    proceeding.append(node)
    if node in graph:
        for successor in graph[node]:
            if component_has_cycle(successor, graph, proceeding, visited):
                return True
    proceeding.remove(node)
    visited.add(node)
    return False

def has_cycle(graph):
    visited = set()
    proceeding = list()
    todo = set(graph.keys())

    while todo:
        node = todo.pop()
        if component_has_cycle(node, graph, proceeding, visited):
            i = proceeding.index(proceeding[-1])
            return proceeding[i:]
        todo.difference_update(visited)
    return []

class CyclicDependencies(Exception):
    def __init__(self, seq):
        self.seq = seq

    def __str__(self):
        return "cyclic dependencies: {}".format(" -> ".join(self.seq))

def remove_duplicates(l):
    d = {}
    nl = []
    for e in l:
        if not e in d:
            d[e] = 1
            nl.append(e)
    return nl

class CodeGen (asdl.VisitorBase) :

    def __init__(self, node):
        asdl.VisitorBase.__init__(self)

        self.starting_node = None
        self.current_node = None
        self.hierarchy = defaultdict(list)
        self.hierarchy['_AST'] = []
        self.fields = defaultdict(list)
        self.attributes = defaultdict(list)
        self.code = defaultdict(list)

        self.visit(node)
        ret = has_cycle(self.hierarchy)
        if ret:
            raise CyclicDependencies(ret)
        self._gen_code(node)

    def visitModule(self, node):
        for name, child in node.types.items():
            if not self.starting_node:
                self.starting_node = str(name)
            self.current_node = str(name)
            self.hierarchy[name]
            self.visit(child)

    def visitSum(self, node):
        if hasattr(node, "fields"):
            self.fields[self.current_node] = node.fields
        else:
            self.fields[self.current_node] = []
        if hasattr(node, "attributes"):
            self.attributes[self.current_node] = node.attributes
        else:
            self.attributes[self.current_node] = []

        for child in node.types:
            self.visit(child)

    def visitConstructor (self, node):
        if str(node.name) in self.fields:
            #print >> sys.stderr, "constructor '{!s}' appears twice !".format(node.name)
            #exit(0)
            return
        self.fields[str(node.name)].extend(node.fields)
        self.hierarchy[str(node.name)].append(self.current_node)

    def visitProduct(self, node):
        self.fields[self.current_node].extend(node.fields)

    @memoize
    def _get_fields(self, name):
        if self.fields.has_key(name):
            fields = map(lambda f : f.name, self.fields[name])
            for parent in self.hierarchy[name]:
                fields.extend(self._get_fields(parent))
            return fields
        else:
            return []

    @memoize
    def _get_attributes(self, name):
        if name == '_AST':
            return []
        attributes = map(lambda a : a.name, self.attributes[name])
        for parent in self.hierarchy[name]:
            attributes.extend(self._get_attributes(parent))
        return attributes

    def _gen_code(self, node):
        is_methods = []
        for name in sorted(self.hierarchy):
            if name != '_AST':
                is_methods.extend(["",
                                   "def is{!s}(self):".format(name),
                                   ["return False"]
                                  ])
        cls = ["class _AST (ast.AST):",
               ["_fields = ()",
                "_attributes = ()",
                "",
                "def __init__ (self, **ARGS):",
                ["ast.AST.__init__(self)",
                 "for k, v in ARGS.items():",
                 ["setattr(self, k, v)"]
                ]
               ] + is_methods
              ]
        self.code['_AST'] = cls

        for name, parents in self.hierarchy.iteritems():
            if name == '_AST':
                continue
            if not parents:
                parents = ['_AST']
            fields = self.fields[name]
            args = []
            assign = []
            body = []
            _fields = remove_duplicates(self._get_fields(name))
            _attributes = remove_duplicates(self._get_attributes(name))

            body = []
            cls = ["class {!s} ({!s}):".format(name, ", ".join(parents)), body]

            non_default_args = []
            default_args = []
            for f in fields:
                if f.name.value == 'ctx':
                    f.opt = True

                if f.opt:
                    default_args.append("{!s}=None".format(f.name))
                    assign.append("self.{0!s} = {0!s}".format(f.name))
                elif f.seq:
                    default_args.append("{!s}=[]".format(f.name))
                    assign.append("self.{0!s} = list({0!s})".format(f.name))
                else:
                    non_default_args.append("{!s}".format(f.name))
                    assign.append("self.{0!s} = {0!s}".format(f.name))

            args = non_default_args + default_args

            body.append("_fields = {!r}".format( tuple(map(repr, _fields))))
            body.append("_attributes = {!r}".format( tuple(map(repr, _attributes))))
            body.append("")
            # ctor
            args_str = ", ".join(args)
            if args_str != "":
                args_str += ", "
            body.append("def __init__ (self, {!s} **ARGS):".format(args_str))
            ctor_body = []
            body.append(ctor_body)
            ctor_body.extend(map(lambda base : "{!s}.__init__(self, **ARGS)".format(base), parents))
            ctor_body.extend(assign)

            body.extend(["", "def is{}(self):".format(name), ["return True"]])

            self.code[name] = cls

    @memoize
    def _cost(self, name):
        # print "call cost {}".format(name)
        if name == '_AST':
            return 0
        parents = self.hierarchy[name]
        return reduce(lambda acc, x: acc + self._cost(x), parents, 1)

    @property
    def python(self):

        classes = self.hierarchy.keys()
        classes.sort(lambda a, b: self._cost(a) - self._cost(b))

        code = ["from snakes.lang import ast",
                "from ast import *",
                ""]

        for cls in classes:
            code.extend(self.code[cls])
            code.append("")

        def python (code, indent) :
            for line in code :
                if isinstance(line, str) :
                    yield (4*indent) * " " + line
                else :
                    for sub in python(line, indent+1) :
                        yield sub
        return "\n".join(python(code, 0))

def compile_asdl(infilename, outfilename):
    """Helper function to compile asdl files.
    """

    infile = open(infilename, 'r')
    outfile = open(outfilename, 'w')

    scanner = asdl.ASDLScanner()
    parser = asdl.ASDLParser()
    tokens = scanner.tokenize(infile.read())
    node = parser.parse(tokens)

    outfile.write(("# this file has been automatically generated running:\n"
                   "# %s\n# timestamp: %s\n\n") % (" ".join(sys.argv),
                                                   datetime.datetime.now()))
    outfile.write(CodeGen(node).python)
    outfile.close()
    infile.close()

if __name__ == "__main__":
    # a simple CLI
    import getopt
    outfile = sys.stdout
    try :
        opts, args = getopt.getopt(sys.argv[1:], "ho:",
                                   ["help", "output="])
        if ("-h", "") in opts or ("--help", "") in opts :
            opts = [("-h", "")]
            args = [None]
        elif not args :
            raise getopt.GetoptError("no input file provided"
                                     " (try -h to get help)")
        elif len(args) > 1 :
            raise getopt.GetoptError("more than one input file provided")
    except getopt.GetoptError :
        sys.stderr.write("%s: %s\n" % (__file__, sys.exc_info()[1]))
        sys.exit(1)
    for (flag, arg) in opts :
        if flag in ("-h", "--help") :
            print("""usage: %s [OPTIONS] INFILE
    Options:
        -h, --help         print this help and exit
        --output=OUTPUT    set output file""" % __file__)
            sys.exit(0)
        elif flag in ("-o", "--output") :
            outfile = open(arg, "w")
    scanner = asdl.ASDLScanner()
    parser = asdl.ASDLParser()
    tokens = scanner.tokenize(open(args[0]).read())
    node = parser.parse(tokens)
    outfile.write(("# this file has been automatically generated running:\n"
                   "# %s\n# timestamp: %s\n\n") % (" ".join(sys.argv),
                                                   datetime.datetime.now()))
    try:
        outfile.write(CodeGen(node).python)
    except CyclicDependencies as cycle:
        msg = "[E] {!s}".format(cycle)
        outfile.write(msg)
        if outfile != sys.stdout:
            print >> sys.stderr, msg
    outfile.close()



"""A Python implementation of CPython's parser

This module is largely based on Jonathan Riehl's PyPgen included in
the Basil framework (http://code.google.com/p/basil).
"""

from snakes import SnakesError
from snakes.lang import ast
import tokenize, string, pprint, warnings, inspect, os.path
from snakes.compat import *

def warn (message) :
    """Issue a warning message.
    """
    warnings.warn(message, stacklevel=2)

class Token (str) :
    """A token from the lexer.

    Behaves as a string that is either the token value (if not empty),
    or the token name. Additional attributes allow to extract various
    information:

     - self.kind: token number (like tokenize.ENDMARKER)
       also available as int(self)
     - self.text: token text
     - self.srow: start row
     - self.scol: start column
     - self.erow: end row
     - self.ecol: end column
     - self.line: full line of text from which the token comes
     - self.name: token name (ie, tokenize.tok_name[self.kind])
     - self.lexer: the Tokenizer instance than produced this token
     - self.filename: input file from which the token comes (or
       '<string>' if None available)
    """
    def __new__ (cls, token, lexer) :
        """Create a new instance.

        __new__ is used instead of __init__ because str is a
        non-mutable object and this __init__ could not assign a str
        content. For more information see:

        http://docs.python.org/reference/datamodel.html#object.__new__
        """
        kind = token[0]
        text = token[1]
        name = lexer.tok_name[kind]
        self = str.__new__(cls, text or name)
        self.kind = kind
        self.text = text
        self.srow, self.scol = token[2]
        self.erow, self.ecol = token[3]
        self.line = token[4]
        self.name = name
        self.lexer = lexer
        try :
            self.filename = lexer.infile.name
        except :
            self.filename = "<string>"
        return self
    def __int__ (self) :
        """Coercion to int (return self.kind).
        """
        return self.kind

class Location (str) :
    """A position in a parsed file

    Used to aggregate the positions of all the tokens is a parse
    (sub)tree. The following attributes are available:

     - self.srow: start row
     - self.scol: start column
     - self.erow: end row
     - self.ecol: end column
     - self.filename: input file from which the token comes (or
       '<string>' if None available)
    """
    def __new__ (cls, first, last) :
        """Create a new instance

        Expected arguments:
         - first: the first Token or Location instance in the region
         - last: the last one
        """
        self = str.__new__(cls, "%s[%s:%s-%s:%s]"
                           % (first.filename, first.srow, first.scol,
                              last.erow, last.ecol))
        self.srow, self.scol = first.srow, first.scol
        self.erow, self.ecol = last.erow, last.ecol
        self.filename = first.filename
        self.lexer = first.lexer
        return self

class ParseError (SnakesError) :
    """Exception raised when parsing fails.

    It's better not to use SyntaxError because this soes not allows to
    distinguish when the text being parser has a syntax error from
    when because the parser itself has a syntax error.
    """
    def __init__ (self, token, expected=None, reason=None) :
        """Initialize a new instance.

        Expected arguments are:
         - token: the erroneous token (a Token instance)
         - expected: either a Token instance or a token kind (like
           tokenize.NAME) to indicated what was expected instead
        """
        self.token = token
        if expected is not None :
            expected = int(expected)
        self.expected = expected
        if token is None :
            pos = ""
        else :
            pos = "%s[%s:%s]: " % (token.filename, token.srow, token.scol)
        if reason is not None :
            msg = reason
        elif self.expected is not None :
            msg = ("expected %s but found %r" %
                   (token.lexer.tok_name[expected], token))
        else :
            msg = "unexpected token %r" % token
        SnakesError.__init__(self, pos + msg)

class Tokenizer (object) :
    """A simple lexical analyser based on Python's tokenize module.

    The differences with tokenize module are:
     - new simple tokens may be added (just strings, no regexps)
     - Python's token may be not included
     - tokens may be automatically skipped (removed from the output)
     - tokenize.OP kind is refined (eg, ':' gets kind tokenize.COLON)

    This class replaces two PyPgen elements:
     - module basil.lang.python.TokenUtils
     - module basil.lang.python.StdTokenizer
    """
    _pyopmap = {
        '(' : tokenize.LPAR,
        ')' : tokenize.RPAR,
        '[' : tokenize.LSQB,
        ']' : tokenize.RSQB,
        ':' : tokenize.COLON,
        ',' : tokenize.COMMA,
        ';' : tokenize.SEMI,
        '+' : tokenize.PLUS,
        '+=' : tokenize.PLUSEQUAL,
        '-' : tokenize.MINUS,
        '-=' : tokenize.MINEQUAL,
        '*' : tokenize.STAR,
        '**' : tokenize.DOUBLESTAR,
        '**=' : tokenize.DOUBLESTAREQUAL,
        '*=' : tokenize.STAREQUAL,
        '/' : tokenize.SLASH,
        '//' : tokenize.DOUBLESLASH,
        '//=' : tokenize.DOUBLESLASHEQUAL,
        '/=' : tokenize.SLASHEQUAL,
        '|' : tokenize.VBAR,
        '|=' : tokenize.VBAREQUAL,
        '&' : tokenize.AMPER,
        '&=' : tokenize.AMPEREQUAL,
        '<' : tokenize.LESS,
        '<=' : tokenize.LESSEQUAL,
        '<<' : tokenize.LEFTSHIFT,
        '<<=' : tokenize.LEFTSHIFTEQUAL,
        '>' : tokenize.GREATER,
        '>=' : tokenize.GREATEREQUAL,
        '>>' : tokenize.RIGHTSHIFT,
        '>>=' : tokenize.RIGHTSHIFTEQUAL,
        '=' : tokenize.EQUAL,
        '==' : tokenize.EQEQUAL,
        '.' : tokenize.DOT,
        '%' : tokenize.PERCENT,
        '%=' : tokenize.PERCENTEQUAL,
        '{' : tokenize.LBRACE,
        '}' : tokenize.RBRACE,
        '^' : tokenize.CIRCUMFLEX,
        '^=' : tokenize.CIRCUMFLEXEQUAL,
        '~' : tokenize.TILDE,
        '!=' : tokenize.NOTEQUAL,
        '<>' : tokenize.NOTEQUAL,
        '@' : tokenize.AT
        }
    def __init__ (self, python=True, opmap={}, skip=None, **extra) :
        """Initialize a new instance.

        Expected arguments are:
         - python: a bool to indicate whether to include or not
           Python's tokens (default to True)
         - opmap: a dict to map litteral tokens (given as '...' in the
           grammar) to token kinds (default to {}). This parameter is
           useful only to redefine Python's mapping
         - skip: a collection of tokens that the tokenizer will
           automatically skip (default to [COMMENT, NL])
         - additional keywords arguments allow to define new tokens,
           for instance, providing DOLLAR='$' defines a new token
           called 'DOLLAR' (its kind will be automatically computed)

        An instance of Tokenizer has the following attributes:
         - self.opmap: a dict mapping operators token literals to the
           corresponding kind, for instance, ':' is mapped to
           tokenize.COLON (this can be overridden using argument
           opmap)
         - self.tok_name: a replacement of tokenize.tok_name that also
           include the user-defined tokens
         - for each token called FOO (including user-defined ones), an
           attribute self.FOO hols the corresponding kind
        """
        self._python = python
        self._opmap = opmap.copy()
        if python :
            self.opmap = self._pyopmap.copy()
            self.opmap.update(opmap)
        else :
            self.opmap = opmap.copy()
        self.tok_name = {}
        self._extra = {}
        if python :
            for kind, name in tokenize.tok_name.items() :
                self.tok_name[kind] = name
                setattr(self, name, kind)
        if not hasattr(self, "NT_OFFSET") :
            self.NT_OFFSET = 256
        last = max(n for n in self.tok_name if n != self.NT_OFFSET)
        for shift, (name, txt) in enumerate(sorted(extra.items())) :
            #WARNING: sorted above is required to guaranty that extra
            # tokens will always get the same number (dict order is
            # not guaranteed)
            kind = last + shift
            if kind >= self.NT_OFFSET :
                raise TypeError("too many new tokens")
            self.tok_name[kind] = name
            setattr(self, name, kind)
            self._extra[txt] = kind
        self.opmap.update(self._extra)
        if skip is None :
            skip = [self.COMMENT, self.NL]
        self._skip = set(skip)
    def __repr__ (self) :
        """Encodes an instance as Python source code.

        Non-default arguments provided to the constructor are included
        so that exactly the same Tokenizer instance can be recovered
        from the returned source code.

        >>> print repr(Tokenizer())
        Tokenizer()
        >>> print repr(Tokenizer(DOLLAR='$'))
        Tokenizer(DOLLAR='$')
        >>> print repr(Tokenizer(skip=[], DOLLAR='$'))
        Tokenizer(skip=[], DOLLAR='$')
        """
        args = []
        if not self._python :
            args.append("python=%s" % self._python)
        if self._opmap :
            args.append("opmap=%r" % self._opmap)
        if self._skip != set([self.COMMENT, self.NL]) :
            args.append("skip=%r" % list(self._skip))
        args.extend("%s=%r" % (self.tok_name[kind], txt) for txt, kind
                    in self._extra.items())
        return "%s(%s)" % (self.__class__.__name__, ", ".join(args))
    def tokenize (self, stream) :
        """Break an input stream into tokens.

        Expected argument is:
         - stream: a file-like object (with a method readline)

        Return a generator of Token instances, ParseError is raised
        whenever an erroneous token is encountered.

        This is basically the same as tokenize.generate_tokens but:
         - the appropriate tokens are skipped
         - OP kind is converted according to self.opmap
         - user-defined tokens are handled

        During the iteration, two more attributes can be used:
         - self.last: last recognized token (ie, last yielded)
         - self.infile: the input stream passed to method tokenize
        """
        self.infile = stream
        self.last = None
        self.lines = []
        def readline () :
            self.lines.append(stream.readline())
            return self.lines[-1]
        err = self.ERRORTOKEN
        for token in tokenize.generate_tokens(readline) :
            if token[0] == err :
                try :
                    token = (self._extra[token[1]],) + token[1:]
                except :
                    raise ParseError(Token(token, self))
            elif token[0] in self._skip :
                try:
                    self.skip_token(Token(token, self))
                except :
                    pass
                continue
            elif token[0] == self.OP :
                token = (self.opmap[token[1]],) + token[1:]
            self.last = Token(token, self)
            yield self.last
    def skip_token (self, token) :
        pass

try :
    Tokenizer._pyopmap['`'] = tokenize.BACKQUOTE
except AttributeError :
    pass

class PgenParser (object) :
    """A parser for pgen files.

    The following grammar is used:

        mstart : ( rule | NEWLINE | newtok )* ENDMARKER
        newtok : '$' NAME STRING NEWLINE
        rule : NAME COLON rhs NEWLINE
        rhs := alt ( VBAR alt )*
        alt : item+
        item : LSQB rhs RSQB | atom ( STAR | PLUS )?
        atom : LPAR rhs RPAR | NAME | STRING

    With respect to PyPgen, an additional rule 'newtok' has been added
    to allow for user-defined tokens.

    This class is adapted from module basil.parsing.PgenParser, it has
    attributes MSTART, ..., provinding to the symbol numbers for the
    corresponding rules.
    """
    MSTART = 256
    RULE = 257
    RHS = 258
    ALT = 259
    ITEM = 260
    ATOM = 261
    NEWTOK = 262
    def __init__ (self) :
        self.lexer = Tokenizer(NEWTOK='$')
    def expect (self, expected, found) :
        if expected != found.kind :
            raise ParseError(found, expected=expected)
    @classmethod
    def parse (cls, filename) :
        """Parse a pgen file.

        Expected argument is:
         - filename: path of pgen file to be parsed

        Return a 2-tuple (G, T) where:
         - G is the grammar's syntax tree
         - T is a Tokenizer instance to be used with the parser
           generated from this grammar (ie, including all the required
           user-defined tokens)

        This is basically the only method that is needed:

        >>> mygrammar = PgenParser.parse('myfile.pgen')
        """
        return cls().parse_file(filename)
    def parse_file (self, filename) :
        """Parse a pgen file.

        Expected argument is:
         - filename: path of pgen file to be parsed

        Return a 2-tuple (g, t):
         - g: is the grammar's syntax tree
         - t: is a Tokenizer instance to be used with the parser
           generated from this grammar (ie, including all the required
           user-defined tokens)

        Recognize mstart : ( rule | NEWLINE )* ENDMARKER

        Like in PyPgen, the parser is recursive descendent, each rule
        'X' being recognized by a dedicated method 'handleX' (except
        for 'mstart'). Each such method expects a current token (or
        None if it has to be fetched from the tokenizer) and returns a
        2-tuple (R, T) where:
         - R is the resulting syntax tree
         - T is the new current token (or None)
        """
        self.infile = open(filename)
        self.tokens = self.lexer.tokenize(self.infile)
        extra = {}
        children = []
        current = next(self.tokens)
        while current.kind != self.lexer.ENDMARKER :
            if current.kind == self.lexer.NEWLINE :
                children.append((current, []))
                current = None
            elif current.kind == self.lexer.NEWTOK :
                name, text = self.handleNewtok(current)
                current = None
                extra[name] = text
            else :
                ruleResult, current = self.handleRule(current)
                children.append(ruleResult)
            if current is None :
                current = next(self.tokens)
        children.append((current, []))
        return (self.MSTART, children), Tokenizer(**extra)
    def handleNewtok (self, current=None) :
        """Recognize newtok : '$' NAME STRING NEWLINE

        Unlike the other 'handleX' methods, this one does not return a
        syntax tree because it implements a parsing directive.
        Instead, it returns a 2-tuple (N, S) where:
         - N is the user-defined token name
         - S is the token string value
        """
        if current is None :
            current = next(self.tokens)
        self.expect(self.lexer.NEWTOK, current)
        name = next(self.tokens)
        self.expect(self.lexer.NAME, name)
        text = next(self.tokens)
        self.expect(self.lexer.STRING, text)
        nl = next(self.tokens)
        self.expect(self.lexer.NEWLINE, nl)
        return name, compile(text, "<string>", "eval",
                             ast.PyCF_ONLY_AST).body.s
    def handleRule (self, current=None) :
        """Recognize rule : NAME COLON rhs NEWLINE
        """
        children = []
        if current is None :
            current = next(self.tokens)
        self.expect(self.lexer.NAME, current)
        children.append((current, []))
        current = next(self.tokens)
        self.expect(self.lexer.COLON, current)
        children.append((current, []))
        rhsResult, current = self.handleRhs()
        children.append(rhsResult)
        if current is None :
            current = next(self.tokens)
        self.expect(self.lexer.NEWLINE, current)
        children.append((current, []))
        result = (self.RULE, children)
        return result, None
    def handleRhs (self, current=None) :
        """Recognize rhs : alt ( VBAR alt )*
        """
        children = []
        altResult, current = self.handleAlt(current)
        children.append(altResult)
        if current is None :
            current = next(self.tokens)
        while current.kind == self.lexer.VBAR :
            children.append((current, []))
            altResult, current = self.handleAlt()
            children.append(altResult)
            if current is None :
                current = next(self.tokens)
        result = (self.RHS, children)
        return result, current
    def handleAlt (self, current=None) :
        """ Recognize alt : item+
        """
        children = []
        itemResult, current = self.handleItem(current)
        children.append(itemResult)
        if current is None :
            current = next(self.tokens)
        while current.kind in (self.lexer.LSQB, self.lexer.LPAR,
                               self.lexer.NAME, self.lexer.STRING) :
            itemResult, current = self.handleItem(current)
            children.append(itemResult)
            if current is None :
                current = next(self.tokens)
        return (self.ALT, children), current
    def handleItem (self, current=None) :
        """Recognize item : LSQB rhs RSQB | atom ( STAR | PLUS )?
        """
        children = []
        if current is None :
            current = next(self.tokens)
        if current.kind == self.lexer.LSQB :
            children.append((current, []))
            rhsResult, current = self.handleRhs()
            children.append(rhsResult)
            if current is None :
                current = next(self.tokens)
            self.expect(self.lexer.RSQB, current)
            children.append((current, []))
            current = None
        else :
            atomResult, current = self.handleAtom(current)
            children.append(atomResult)
            if current is None :
                current = next(self.tokens)
            if current.kind in (self.lexer.STAR, self.lexer.PLUS) :
                children.append((current, []))
                current = None
        return (self.ITEM, children), current
    def handleAtom (self, current=None) :
        """Recognize atom : LPAR rhs RPAR | NAME | STRING
        """
        children = []
        if current is None :
            current = next(self.tokens)
        tokType = current.kind
        if tokType == self.lexer.LPAR :
            children.append((current, []))
            rhsResult, current = self.handleRhs()
            children.append(rhsResult)
            if current is None :
                current = next(self.tokens)
            self.expect(self.lexer.RPAR, current)
            children.append((current, []))
        elif tokType == self.lexer.STRING :
            children.append((current, []))
        else :
            self.expect(self.lexer.NAME, current)
            children.append((current, []))
        return (self.ATOM, children), None

class Parser (object) :
    """A LL1 parser for a generated grammar.

    This class aggregates two elements from PyPgen:
     - module basil.lang.python.DFAParser
     - class basil.parsing.PyPgen.PyPgenParser

    The main differences are:
     - simplified interface
     - adapt to handle Token instances instead of 3-tuples (kind,
       text, lineno)
     - remove functions arguments that can now be retreived as
       instance attributes
     - use Python warnings instead of print statements
     - many minor code edits (for my own understanding)

    Docstrings are provided only for methods that did not have an
    equivalent in PyPgen or that have been changed in a significant
    way.
    """
    def __init__ (self, grammar, tokenizer) :
        """Initialize a new instance.

        Expected arguments are:
         - grammar: the gramar object as returned by PyPgen.grammar()
         - tokenizer: a Tokenizer instance suitable for this grammar
           (eg, that passed to PyPgen's constructor)
        """
        self.grammar = grammar
        self.start = grammar[2]
        self.stringMap = {}
        for dfa in self.grammar[0] :
            dfaType, dfaName = dfa[:2]
            self.stringMap[dfaName] = dfaType
        self.symbolMap = {}
        for dfa in self.grammar[0] :
            dfaType, dfaName = dfa[:2]
            self.symbolMap[dfaType] = dfaName
        self.tokenizer = tokenizer
        self.addAccelerators()
    def parseTokens (self, tokens, start=None) :
        """Parse a series of tokens.

        Expected arguments:
         - tokens: a generator of Token instances
         - start: the start symbol to be recognized (or None to use
           the default one)

        The token generator should provide only tokens that are
        compatible with the tokenizer passed to the Parser's
        constructor. Otherwise, ParseError will be raised as unknown
        tokens will be issued.
        """
        self.tokens = tokens
        return self._parse(start)
    def parseFile (self, filename, start=None) :
        """Parse a text file provided by its path.

        Expected arguments:
         - filename: a file name
         - start: the start symbol to be recognized (or None to use
           the default one)

        The start symbol may be provided by its number (int) or its
        name (str) as specified in the grammar.
        """
        self.tokens = self.tokenizer.tokenize(open(filename))
        return self._parse(start)
    def parseStream (self, stream, start=None) :
        """Parse text from an opened file.

        Expected arguments:
         - stream: a file-like object (with a method readline)
         - start: the start symbol to be recognized (or None to use
           the default one)

        The start symbol may be provided by its number (int) or its
        name (str) as specified in the grammar.
        """
        self.tokens = self.tokenizer.tokenize(stream)
        return self._parse(start)
    def parseString (self, text, start=None, filename="<string>") :
        """Parse text given as a string.

        Expected arguments:
         - text: a string-like object
         - start: the start symbol to be recognized (or None to use
           the default one)

        The start symbol may be provided by its number (int) or its
        name (str) as specified in the grammar.
        """
        data = io.StringIO(text)
        data.name = filename
        self.tokens = self.tokenizer.tokenize(data)
        return self._parse(start)
    def _parse (self, start=None) :
        """Main parsing method.

        Expected argument:
         - start: the start symbol to be recognized (or None to use
           the default one)

        The start symbol may be provided by its number (int) or its
        name (str) as specified in the grammar.
        """
        if start is None :
            start = self.start
        elif start in self.stringMap :
            start = self.stringMap[start]
        elif start not in self.symbolMap :
            raise ValueError("unknown start symbol %r" % start)
        tokens = self.tokens
        # initialize the parsing stack
        rootNode = ((start, None, 0), [])
        dfa = self.findDFA(start)
        self.stack = [(dfa[3][dfa[2]], dfa, rootNode)]
        # parse all of it
        result = self._LL1_OK
        while result == self._LL1_OK :
            result, expected = self.addToken(next(tokens))
        if result == self._LL1_DONE :
            return self._fix_locations(rootNode)
        elif result == self._LL1_SYNTAX :
            raise ParseError(self.tokenizer.last, expected=expected)
    def _tostrings (self, st) :
        """Substitute symbol numbers by strings in a syntax tree.

        Expected argument:
         - st: a syntax tree as returned by the parser
        """
        (kind, token, lineno), children = st
        if kind >= self.tokenizer.NT_OFFSET :
            name = self.symbolMap
        else :
            name = self.tokenizer.tok_name
        return ((name[kind], token, lineno),
                [self._tostrings(c) for c in children])
    def _fix_locations (self, st) :
        """Replaces None in non-terminal nodes by a Location instance.

        Expected argument:
         - st: a syntax tree as returned by the parser
        """
        (kind, token, lineno), children = st
        children = [self._fix_locations(c) for c in children]
        if kind >= self.tokenizer.NT_OFFSET :
            token = Location(children[0][0][1], children[-1][0][1])
        return ((kind, token, lineno), children)
    def pprint (self, st) :
        """Return a human-readable representation of a syntax tree.

        Expected argument:
         - st: a syntax tree as returned by the parser

        All symbol numbers are substituted by the corresponding names
        and the text is indented appropriately.
        """
        return pprint.pformat(self._tostrings(st))
    # the rest of the class has not changed too much
    def addAccelerators (self) :
        if self.grammar[-1] : # already has accelerators
            return
        dfas, labels, start, accel = self.grammar
        def handleState (state) :
            arcs, accel, accept = state
            accept = 0
            labelCount = len(labels)
            accelArray = [-1] * labelCount
            for arc in arcs :
                labelIndex, arrow = arc
                kind = labels[labelIndex][0]
                if arrow >= 128 :
                    warn("too many states (%d >= 128)!" % arrow)
                    continue
                if kind >= self.tokenizer.NT_OFFSET :
                    targetFirstSet = self.findDFA(kind)[4]
                    if kind - self.tokenizer.NT_OFFSET >= 128 :
                        warn("nonterminal too high (%d >= %d)!" %
                             (kind, 128 + self.tokenizer.NT_OFFSET))
                        continue
                    for ibit in range(labelCount) :
                        if self.testbit(targetFirstSet, ibit) :
                            accelVal = (arrow | 128 |
                                        ((kind - self.tokenizer.NT_OFFSET) << 8))
                            oldVal = accelArray[ibit]
                            if oldVal != -1 :
                                # XXX Make this error reporting more better.
                                oldType = oldVal >> 8
                                # FIXME: bug in the original source
                                #warn("ambiguity at bit %d (for %d: was to %x,"
                                #     " now to %x)."
                                #     % (ibit, states.index(state),
                                #        oldVal, accelVal))
                                warn("ambiguity at bit %d" % ibit)
                            accelArray[ibit] = (arrow | 128 |
                                                ((kind - self.tokenizer.NT_OFFSET) << 8))
                elif labelIndex == 0 :
                    accept = 1
                elif labelIndex >= 0 and labelIndex < labelCount :
                    accelArray[labelIndex] = arrow
            # Now compute the upper and lower bounds.
            accelUpper = labelCount
            while accelUpper > 0 and accelArray[accelUpper-1] == -1 :
                accelUpper -= 1
            accelLower = 0
            while accelLower < accelUpper and accelArray[accelLower] == -1 :
                accelLower += 1
            accelArray = accelArray[accelLower:accelUpper]
            return (arcs, (accelUpper, accelLower, accelArray), accept)
        def handleDFA (dfa) :
            kind, name, initial, states, first = dfa
            return (kind, name, initial, list(map(handleState, states)))
        self.grammar = (list(map(handleDFA, dfas)), labels, start, 1)
    _LL1_OK = 0   # replaced E_ prefix with _LL1_ to prevent potential
    _LL1_DONE = 1 # conflicts with grammar symbols
    _LL1_SYNTAX = 2
    def testbit (self, bitstr, ibit) :
        return (ord(bitstr[ibit >> 3]) & (1 << (ibit & 0x7))) != 0
    def classify (self, token) :
        labels = self.grammar[1]
        if token.kind == self.tokenizer.NAME :
            for i, label in enumerate(labels) :
                if (token.kind, token) == label :
                    return i
        for i, label in enumerate(labels) :
            if (token.kind == label[0]) and (label[1] is None) :
                return i
        return -1
    def findDFA (self, start) :
        return self.grammar[0][start - self.tokenizer.NT_OFFSET]
    def addToken (self, token) :
        stack = self.stack
        ilabel = self.classify(token)
        while True :
            state, dfa, parent = stack[-1]
            # Perform accelerator
            arcs, (accelUpper, accelLower, accelTable), accept = state
            if accelLower <= ilabel < accelUpper :
                accelResult = accelTable[ilabel - accelLower]
                if accelResult != -1 :
                    # Handle accelerator result
                    if accelResult & 128 :
                        # Push non-terminal
                        nt = (accelResult >> 8) + self.tokenizer.NT_OFFSET
                        arrow = accelResult & 127
                        nextDFA = self.findDFA(nt)
                        # INLINE PUSH
                        newAstNode = ((nt, None, token.srow), [])
                        parent[1].append(newAstNode)
                        stack[-1] = (dfa[3][arrow], dfa, parent)
                        stack.append((nextDFA[3][nextDFA[2]], nextDFA,
                                      newAstNode))
                        continue
                    # INLINE SHIFT
                    parent[1].append(((token.kind, token, token.srow), []))
                    nextState = dfa[3][accelResult]
                    stack[-1] = (nextState, dfa, parent)
                    state = nextState
                    while state[2] and len(state[0]) == 1 :
                        # INLINE POP
                        stack.pop(-1)
                        if not stack :
                            return self._LL1_DONE, None
                        else :
                            state, dfa, parent = stack[-1]
                    return self._LL1_OK, None
            if accept :
                stack.pop(-1)
                if not stack :
                    return self._LL1_SYNTAX, self.tokenizer.ENDMARKER
                continue
            if ((accelUpper < accelLower) and
                (self.grammar[1][accelLower][1] is not None)) :
                expected = self.grammar[1][accelLower][1]
            else :
                expected = None
            return self._LL1_SYNTAX, expected

class PyPgen (object) :
    """A grammar generator.

    This class aggregates two elements from PyPgen:
     - class basil.parsing.PyPgen.PyPgen
     - function basil.parsing.PyPgen.buildParser
     - parts of function basil.parsing.PyPgen.parserMain

    The main differences are:
     - simplified interface
     - adapt to handle Token instances instead of 3-tuples (kind,
       text, lineno)
     - remove functions arguments that can now be retreived as
       instance attributes
     - use Python warnings instead of print statements
     - many minor code edits (for my own understanding)

    Docstrings are provided only for methods that did not have an
    equivalent in PyPgen or that have been changed in a significant
    way.
    """
    def __init__ (self, gst, tokenizer) :
        """Initialize a new instance.

        Expected arguments are:
         - gst: the grammar's syntax tree as returned by
           PgenParser.parse()
         - tokenizer: a Tokenizer instance suitable for this grammar,
           also returned by PgenParser.parse()
        """
        self.tokenizer = tokenizer
        self.EMPTY = self.tokenizer.ENDMARKER
        self.gst = gst
        self.nfaGrammar = self.dfaGrammar = None
        self.nfa = None
        self.crntKind = self.tokenizer.NT_OFFSET
        self.operatorMap = tokenizer.opmap
    def grammar (self) :
        """Generate and return the grammar object.
        """
        nfaGrammar = self.handleStart(self.gst)
        grammar = self.generateDfaGrammar(nfaGrammar)
        self.translateLabels(grammar)
        self.generateFirstSets(grammar)
        grammar[0] = list(map(tuple, grammar[0]))
        # Trick to add accelerators at generation time: it's easier to
        # do it this way than to extract the required elements from
        # class Parser.
        return Parser(tuple(grammar), self.tokenizer).grammar
    def python (self, pgen="pgen", inline=False) :
        """Build and return Python code for parsing module.

        Expected arguments are:
         - pgen: the name of module pgen in the generated source
           (default to 'pgen')
         - inline: a bool value to indicate whether to import or
           inline pgen module in the generated code

        If inline=True, the generated code is much bigger but does not
        depend on any non-standard module.
        """
        pysrc = ("%(pgen)s"
                 "tokenizer = %(prefix)s%(tokenizer)r\n"
                 "grammar = %(grammar)s\n"
                 "parser = %(prefix)sParser(grammar, tokenizer)\n\n"
                 "if __name__ == '__main__' :\n"
                 "    # just for test purpose\n"
                 "    import sys, pprint\n"
                 "    st = parser.parseStream(sys.stdin)\n"
                 "    print(parser.pprint(st))\n")
        format = {"grammar" : pprint.pformat(self.grammar()),
                  "prefix" : pgen + ".",
                  "tokenizer" : self.tokenizer,
                  "pgen" : "import tokenize, %s\n\n" % pgen,
                  "inline" : "",
                  }
        if inline :
            format["prefix"] = ""
            source = inspect.getsource(inspect.getmodule(self))
            source = source.rsplit("if __name__ == '__main__' :", 1)[0]
            format["pgen"] = ("### module '%s.py' inlined\n"
                              "%s\n### end of '%s.py'\n\n"
                              % (pgen, source.rstrip(), pgen))
        return pysrc % format
    @classmethod
    def translate (cls, src, tgt=None, pgen="pgen", inline=False) :
        """Translate a pgen file to a Python file that implements the
        corresponding parser.

        Expected arguments are:
         - src: path of the pgen file
         - tgt: path of target Python file, if None, its name is
           derived from src (replacing its extension by .py)
         - pgen, inline: like in PyPgen.python()

        Warning: the output file is silently overwritten if it already
        exist.
        """
        if tgt is None :
            tgt = os.path.splitext(src)[0] + ".py"
        gst, tokenizer = PgenParser.parse(src)
        self = PyPgen(gst, tokenizer)
        outfile = open(tgt, "w")
        outfile.write(("# this file has been automatically generated running:\n"
                       "# %s\n\n") % " ".join(sys.argv))
        outfile.write(self.python(pgen, inline))
        outfile.close()
    # the rest of the class has not changed too much
    def addLabel (self, labelList, tokKind, tokName) :
        labelTup = (tokKind, tokName)
        if labelTup in labelList :
            return labelList.index(labelTup)
        labelIndex = len(labelList)
        labelList.append(labelTup)
        return labelIndex
    def handleStart (self, gst) :
        self.nfaGrammar = [[],[(self.tokenizer.ENDMARKER, "EMPTY")]]
        self.crntKind = self.tokenizer.NT_OFFSET
        kind, children = gst
        for child in children :
            if int(child[0]) == PgenParser.RULE :
                self.handleRule(child)
        return self.nfaGrammar
    def handleRule (self, gst) :
        # NFA := [ type : Int, name : String, [ STATE ], start : Int,
        #         finish : Int ]
        # STATE := [ ARC ]
        # ARC := ( labelIndex : Int, stateIndex : Int )
        ####
        # build the NFA shell
        self.nfa = [self.crntKind, None, [], -1, -1]
        self.crntKind += 1
        # work on the AST node
        kind, children = gst
        name, colon, rhs, newline = children
        self.nfa[1] = name[0]
        if (self.tokenizer.NAME, name[0]) not in self.nfaGrammar[1] :
            self.nfaGrammar[1].append((self.tokenizer.NAME, name[0]))
        start, finish = self.handleRhs(rhs)
        self.nfa[3] = start
        self.nfa[4] = finish
        # append the NFA to the grammar
        self.nfaGrammar[0].append(self.nfa)
    def handleRhs (self, gst) :
        kind, children = gst
        start, finish = self.handleAlt(children[0])
        if len(children) > 1 :
            cStart = start
            cFinish = finish
            start = len(self.nfa[2])
            self.nfa[2].append([(self.EMPTY, cStart)])
            finish = len(self.nfa[2])
            self.nfa[2].append([])
            self.nfa[2][cFinish].append((self.EMPTY, finish))
            for child in children[2:] :
                if int(child[0]) == PgenParser.ALT :
                    cStart, cFinish = self.handleAlt(child)
                    self.nfa[2][start].append((self.EMPTY, cStart))
                    self.nfa[2][cFinish].append((self.EMPTY, finish))
        return start, finish
    def handleAlt (self, gst) :
        kind, children = gst
        start, finish = self.handleItem(children[0])
        if len(children) > 1 :
            for child in children[1:] :
                cStart, cFinish = self.handleItem(child)
                self.nfa[2][finish].append((self.EMPTY, cStart))
                finish = cFinish
        return start, finish
    def handleItem (self, gst) :
        nodeKind, children = gst
        if int(children[0][0]) == PgenParser.ATOM :
            start, finish = self.handleAtom(children[0])
            if len(children) > 1 :
                # Short out the child NFA
                self.nfa[2][finish].append((self.EMPTY, start))
                if children[1][0].kind == self.tokenizer.STAR :
                    finish = start
        else :
            start = len(self.nfa[2])
            finish = start + 1
            self.nfa[2].append([(self.EMPTY, finish)])
            self.nfa[2].append([])
            cStart, cFinish = self.handleRhs(children[1])
            self.nfa[2][start].append((self.EMPTY, cStart))
            self.nfa[2][cFinish].append((self.EMPTY, finish))
        return start, finish
    def handleAtom (self, gst) :
        nodeKind, children = gst
        tok = children[0][0]
        if tok.kind == self.tokenizer.LPAR :
            start, finish = self.handleRhs(children[1])
        elif tok.kind in (self.tokenizer.STRING, self.tokenizer.NAME) :
            start = len(self.nfa[2])
            finish = start + 1
            labelIndex = self.addLabel(self.nfaGrammar[1], tok.kind, tok)
            self.nfa[2].append([(labelIndex, finish)])
            self.nfa[2].append([])
        return start, finish
    def generateDfaGrammar (self, nfaGrammar, start=None) :
        # See notes in pgen.lang.python.DFAParser for output schema.
        dfas = []
        for nfa in nfaGrammar[0] :
            dfas.append(self.nfaToDfa(nfa))
        kind = dfas[0][0]
        if start is not None :
            found = False
            for dfa in dfas :
                if dfa[1] == start :
                    kind = dfa[0]
                    found = True
                    break
            if not found :
                warn("couldn't find nonterminal %r, "
                     "using %r instead." % (start, dfas[0][1]))
        return [dfas, nfaGrammar[1][:], kind, 0]
    def addClosure (self, stateList, nfa, istate) :
        stateList[istate] = True
        arcs = nfa[2][istate]
        for label, arrow in arcs :
            if label == self.EMPTY :
                self.addClosure(stateList, nfa, arrow)
    def nfaToDfa (self, nfa) :
        tempStates = []
        crntTempState = [[False] * len(nfa[2]), [], False]
        self.addClosure(crntTempState[0], nfa, nfa[3])
        crntTempState[2] = crntTempState[0][nfa[4]]
        if crntTempState[2] :
            warn("nonterminal %r may produce empty." % nfa[1])
        tempStates.append(crntTempState)
        index = 0
        while index < len(tempStates) :
            crntTempState = tempStates[index]
            for componentState in range(len(nfa[2])) :
                if not crntTempState[0][componentState] :
                    continue
                nfaArcs = nfa[2][componentState]
                for label, nfaArrow in nfaArcs :
                    if label == self.EMPTY :
                        continue
                    foundTempArc = False
                    for tempArc in crntTempState[1] :
                        if tempArc[0] == label :
                            foundTempArc = True
                            break
                    if not foundTempArc :
                        tempArc = [label, -1, [False] * len(nfa[2])]
                        crntTempState[1].append(tempArc)
                    self.addClosure(tempArc[2], nfa, nfaArrow)
            for arcIndex in range(len(crntTempState[1])) :
                label, arrow, targetStateList = crntTempState[1][arcIndex]
                targetFound = False
                arrow = 0
                for destTempState in tempStates :
                    if targetStateList == destTempState[0] :
                        targetFound = True
                        break
                    arrow += 1
                if not targetFound :
                    assert arrow == len(tempStates)
                    tempState = [targetStateList[:], [],
                                 targetStateList[nfa[4]]]
                    tempStates.append(tempState)
                # Write arrow value back to the arc
                crntTempState[1][arcIndex][1] = arrow
            index += 1
        tempStates = self.simplifyTempDfa(nfa, tempStates)
        return self.tempDfaToDfa(nfa, tempStates)
    def sameState (self, s1, s2) :
        if len(s1[1]) != len(s2[1]) or s1[2] != s2[2] :
            return False
        for arcIndex in range(len(s1[1])) :
            arc1 = s1[1][arcIndex]
            arc2 = s2[1][arcIndex]
            if arc1[:-1] != arc2[:-1] :
                return False
        return True
    def simplifyTempDfa (self, nfa, tempStates) :
        changes = True
        deletedStates = []
        while changes :
            changes = False
            for i in range(1, len(tempStates)) :
                if i in deletedStates :
                    continue
                for j in range(i) :
                    if j in deletedStates :
                        continue
                    if self.sameState(tempStates[i], tempStates[j]) :
                        deletedStates.append(i)
                        for k in range(len(tempStates)) :
                            if k in deletedStates :
                                continue
                            for arc in tempStates[k][1] :
                                if arc[1] == i :
                                    arc[1] = j
                        changes = True
                        break
        for stateIndex in deletedStates :
            tempStates[stateIndex] = None
        return tempStates
    def tempDfaToDfa (self, nfa, tempStates) :
        dfaStates = []
        dfa = [nfa[0], nfa[1], 0, dfaStates, None]
        stateMap = {}
        tempIndex = 0
        for tempState in tempStates :
            if tempState is not None :
                stateMap[tempIndex] = len(dfaStates)
                dfaStates.append(([], (0,0,()), 0))
            tempIndex += 1
        for tempIndex in stateMap.keys() :
            stateList, tempArcs, accepting = tempStates[tempIndex]
            dfaStateIndex = stateMap[tempIndex]
            dfaState = dfaStates[dfaStateIndex]
            for tempArc in tempArcs :
                dfaState[0].append((tempArc[0], stateMap[tempArc[1]]))
            if accepting :
                dfaState[0].append((self.EMPTY, dfaStateIndex))
        return dfa
    def translateLabels (self, grammar) :
        tokenNames = list(self.tokenizer.tok_name.values())
        # Recipe 252143 (remixed for laziness)
        tokenValues = dict(([v, k] for k, v in
                            self.tokenizer.tok_name.items()))
        labelList = grammar[1]
        for labelIndex, (kind, name) in enumerate(labelList) :
            if kind == self.tokenizer.NAME :
                isNonTerminal = False
                for dfa in grammar[0] :
                    if dfa[1] == name :
                        labelList[labelIndex] = (dfa[0], None)
                        isNonTerminal = True
                        break
                if not isNonTerminal :
                    if name in tokenNames :
                        labelList[labelIndex] = (tokenValues[name], None)
                    else :
                        warn("can't translate NAME label '%s'" % name)
            elif kind == self.tokenizer.STRING :
                assert name[0] == name[-1]
                sname = name[1:-1]
                if (sname[0] in string.letters) or (sname[0] == "_") :
                    labelList[labelIndex] = (self.tokenizer.NAME, sname)
                elif sname in self.operatorMap :
                    labelList[labelIndex] = (self.operatorMap[sname],
                                             None)
                else :
                    warn("can't translate STRING label %s" % name)
        return grammar
    def calcFirstSet (self, grammar, dfa) :
        if dfa[4] == -1 :
            warn("left-recursion for %r" % dfa[1])
            return
        if dfa[4] != None :
            warn("re-calculating FIRST set for %r" % dfa[1])
        dfa[4] = -1
        symbols = []
        result = 0
        state = dfa[3][dfa[2]]
        for arc in state[0] :
            sym = arc[0]
            if sym not in symbols :
                symbols.append(sym)
                kind = grammar[1][sym][0]
                if kind >= self.tokenizer.NT_OFFSET :
                    # Nonterminal
                    ddfa = grammar[0][kind - self.tokenizer.NT_OFFSET]
                    if ddfa[4] == -1 :
                        warn("left recursion below %r" % dfa[1])
                    else :
                        if ddfa[4] == None :
                            self.calcFirstSet(grammar, ddfa)
                        result |= ddfa[4]
                else :
                    result |= 1 << sym
        dfa[4] = result
    def generateFirstSets (self, grammar) :
        dfas = grammar[0]
        index = 0
        while index < len(dfas) :
            dfa = dfas[index]
            if None == dfa[4] :
                self.calcFirstSet(grammar, dfa)
            index += 1
        for dfa in dfas :
            set = dfa[4]
            result = []
            while set > 0 :
                crntBits = set & 0xff
                result.append(chr(crntBits))
                set >>= 8
            properSize = (len(grammar[1]) / 8) + 1
            if len(result) < properSize :
                result.append('\x00' * (properSize - len(result)))
            dfa[4] = "".join(result)
        return grammar

if __name__ == '__main__' :
    # a simple CLI
    import sys, getopt
    tgt, pgen, inline = None, "snakes.lang.pgen", False
    try :
        opts, args = getopt.getopt(sys.argv[1:], "h",
                                   ["help", "inline", "output=",
                                    "pgen=", "start="])
        if ("-h", "") in opts or ("--help", "") in opts :
            opts = [("-h", "")]
            args = [None]
        elif not args :
            raise getopt.GetoptError("no input file provided"
                                     " (try -h to get help)")
        elif len(args) > 1 :
            raise getopt.GetoptError("more than one input file provided")
    except getopt.GetoptError :
        sys.stderr.write("%s: %s\n" % (__file__, sys.exc_info()[1]))
        sys.exit(1)
    for (flag, arg) in opts :
        if flag in ("-h", "--help") :
            print("""usage: %s [OPTIONS] INFILE
    Options:
        -h, --help         print this help and exit
        --inline           inline 'pgen.py' in the generated file
        --output=OUTPUT    set output file
        --pgen=PGEN        name of 'pgen' module in output file""" % __file__)
            sys.exit(0)
        elif flag == "--inline" :
            inline = True
        elif flag == "--output" :
            tgt = arg
        elif flag == "--pgen" :
            pgen = arg
    PyPgen.translate(args[0], tgt=tgt, pgen=pgen, inline=inline)



"""A plugin to add labels to nodes and nets. Labels are names (valid
Python identifiers) associated to arbitrary objects.

>>> import snakes.plugins
>>> snakes.plugins.load('labels', 'snakes.nets', 'nets')
<module ...>
>>> from nets import *
>>> t = Transition('t')
>>> t.label(foo='bar', spam=42)
>>> t.label('foo')
'bar'
>>> t.label('spam')
42

Note that when nodes in a Petri net are merged, their labels are
merged too, but in an arbitrary order. So, for example:

>>> n = PetriNet('N')
>>> n.add_place(Place('p1'))
>>> n.place('p1').label(foo='bar', spam='ham')
>>> n.add_place(Place('p2'))
>>> n.place('p2').label(hello='world', spam='egg')
>>> n.merge_places('p', ['p1', 'p2'])
>>> n.place('p').label('hello')
'world'
>>> n.place('p').label('foo')
'bar'
>>> n.place('p').label('spam') in ['ham', 'egg']
True

In the latter statement, we cannot know whether the label will be one
or the other value because merging has been done in an arbitrary
order.
"""

from snakes.plugins import plugin, new_instance
from snakes.pnml import Tree

@plugin("snakes.nets")
def extend (module) :
    class Transition (module.Transition) :
        def label (self, *get, **set) :
            """Get and set labels for the transition. The labels given
            in `get` will be returned as a `tuple` and the labels
            assigned in `set` will be changed accordingly. If a label
            is given both in `get`and `set`, the returned value is
            that it had at the beginning of the call, ie, before it is
            set by the call.

            @param get: labels which values have to be returned
            @type get: `str`
            @param set: labels which values have to be changed
            @type set: `object`
            @return: the tuples of values corresponding to `get`
            @rtype: `tuple`
            @raise KeyError: when a label given in `get` has not been
                assigned previouly
            """
            if not hasattr(self, "_labels") :
                self._labels = {}
            result = tuple(self._labels[g] for g in get)
            self._labels.update(set)
            if len(get) == 1 :
                return result[0]
            elif len(get) > 1 :
                return result
            elif len(set) == 0 :
                return self._labels.copy()
        def has_label (self, name, *names) :
            """Check is a label has been assigned to the transition.

            @param name: the label to check
            @type name: `str`
            @param names: additional labels to check, if used, the
                return value is a `tuple` of `bool` instead of a
                single `bool`
            @return: a Boolean indicating of the checked labels are
                present or not in the transitions
            @rtype: `bool`
            """
            if len(names) == 0 :
                return name in self._labels
            else :
                return tuple(n in self._labels for n in (name,) + names)
        # apidoc stop
        def copy (self, name=None, **options) :
            if not hasattr(self, "_labels") :
                self._labels = {}
            result = module.Transition.copy(self, name, **options)
            result._labels = self._labels.copy()
            return result
        def __pnmldump__ (self) :
            """
            >>> t = Transition('t')
            >>> t.label(foo='bar', spam=42)
            >>> t.__pnmldump__()
            <?xml version="1.0" encoding="utf-8"?>
            <pnml>
             <transition id="t">
              <label name="foo">
               <object type="str">
                bar
               </object>
              </label>
              <label name="spam">
               <object type="int">
                42
               </object>
              </label>
             </transition>
            </pnml>
            """
            t = module.Transition.__pnmldump__(self)
            if hasattr(self, "_labels") :
                for key, val in self._labels.items() :
                    t.add_child(Tree("label", None,
                                     Tree.from_obj(val),
                                     name=key))
            return t
        @classmethod
        def __pnmlload__ (cls, tree) :
            """
            >>> old = Transition('t')
            >>> old.label(foo='bar', spam=42)
            >>> p = old.__pnmldump__()
            >>> new = Transition.__pnmlload__(p)
            >>> new
            Transition('t', Expression('True'))
            >>> new.__class__
            <class 'snakes.plugins.labels.Transition'>
            >>> new.label('foo', 'spam')
            ('bar', 42)
            """
            t = new_instance(cls, module.Transition.__pnmlload__(tree))
            t._labels = dict((lbl["name"], lbl.child().to_obj())
                             for lbl in tree.get_children("label"))
            return t
    class Place (module.Place) :
        def label (self, *get, **set) :
            "See documentation for `Transition.label` above"
            if not hasattr(self, "_labels") :
                self._labels = {}
            result = tuple(self._labels[g] for g in get)
            self._labels.update(set)
            if len(get) == 1 :
                return result[0]
            elif len(get) > 1 :
                return result
            elif len(set) == 0 :
                return self._labels.copy()
        def has_label (self, name, *names) :
            "See documentation for `Transition.has_label` above"
            if len(names) == 0 :
                return name in self._labels
            else :
                return tuple(n in self._labels for n in (name,) + names)
        # apidoc stop
        def copy (self, name=None, **options) :
            if not hasattr(self, "_labels") :
                self._labels = {}
            result = module.Place.copy(self, name, **options)
            result._labels = self._labels.copy()
            return result
        def __pnmldump__ (self) :
            """
            >>> p = Place('p')
            >>> p.label(foo='bar', spam=42)
            >>> p.__pnmldump__()
            <?xml version="1.0" encoding="utf-8"?>
            <pnml>
             <place id="p">
              <type domain="universal"/>
              <initialMarking>
               <multiset/>
              </initialMarking>
              <label name="foo">
               <object type="str">
                bar
               </object>
              </label>
              <label name="spam">
               <object type="int">
                42
               </object>
              </label>
             </place>
            </pnml>
            """
            t = module.Place.__pnmldump__(self)
            if hasattr(self, "_labels") :
                for key, val in self._labels.items() :
                    t.add_child(Tree("label", None,
                                     Tree.from_obj(val),
                                     name=key))
            return t
        @classmethod
        def __pnmlload__ (cls, tree) :
            """
            >>> old = Place('p')
            >>> old.label(foo='bar', spam=42)
            >>> p = old.__pnmldump__()
            >>> new = Place.__pnmlload__(p)
            >>> new
            Place('p', MultiSet([]), tAll)
            >>> new.__class__
            <class 'snakes.plugins.labels.Place'>
            >>> new.label('foo', 'spam')
            ('bar', 42)
            """
            p = new_instance(cls, module.Place.__pnmlload__(tree))
            p._labels = dict((lbl["name"], lbl.child().to_obj())
                             for lbl in tree.get_children("label"))
            return p
    class PetriNet (module.PetriNet) :
        def label (self, *get, **set) :
            "See documentation for `Transition.label` above"
            if not hasattr(self, "_labels") :
                self._labels = {}
            result = tuple(self._labels[g] for g in get)
            self._labels.update(set)
            if len(get) == 1 :
                return result[0]
            elif len(get) > 1 :
                return result
            elif len(set) == 0 :
                return self._labels.copy()
        def has_label (self, name, *names) :
            "See documentation for `Transition.has_label` above"
            if len(names) == 0 :
                return name in self._labels
            else :
                return tuple(n in self._labels for n in (name,) + names)
        # apidoc stop
        def copy (self, name=None, **options) :
            if not hasattr(self, "_labels") :
                self._labels = {}
            result = module.PetriNet.copy(self, name, **options)
            result._labels = self._labels.copy()
            return result
        def __pnmldump__ (self) :
            """
            >>> n = PetriNet('n')
            >>> n.label(foo='bar', spam=42)
            >>> n.__pnmldump__()
            <?xml version="1.0" encoding="utf-8"?>
            <pnml>
             <net id="n">
              <label name="foo">
               <object type="str">
                bar
               </object>
              </label>
              <label name="spam">
               <object type="int">
                42
               </object>
              </label>
             </net>
            </pnml>
            """
            t = module.PetriNet.__pnmldump__(self)
            if hasattr(self, "_labels") :
                for key, val in self._labels.items() :
                    t.add_child(Tree("label", None,
                                     Tree.from_obj(val),
                                     name=key))
            return t
        @classmethod
        def __pnmlload__ (cls, tree) :
            """
            >>> old = PetriNet('n')
            >>> old.label(foo='bar', spam=42)
            >>> p = old.__pnmldump__()
            >>> new = PetriNet.__pnmlload__(p)
            >>> new
            PetriNet('n')
            >>> new.__class__
            <class 'snakes.plugins.labels.PetriNet'>
            >>> new.label('foo', 'spam')
            ('bar', 42)
            """
            n = new_instance(cls, module.PetriNet.__pnmlload__(tree))
            n._labels = dict((lbl["name"], lbl.child().to_obj())
                             for lbl in tree.get_children("label"))
            return n
        def merge_places (self, target, sources, **options) :
            module.PetriNet.merge_places(self, target, sources, **options)
            new = self.place(target)
            for place in sources :
                new.label(**dict(self.place(place).label()))
        def merge_transitions (self, target, sources, **options) :
            module.PetriNet.merge_transitions(self, target, sources, **options)
            new = self.transition(target)
            for trans in sources :
                new.label(**dict(self.transition(trans).label()))
    return Transition, Place, PetriNet
